from qtpy.QtWidgets import QWidget, QPushButton, QHBoxLayout


class FlowViewZoomWidget(QWidget):
    def __init__(self, flow_view):
        super(FlowViewZoomWidget, self).__init__()

        self.flow_view = flow_view

        self.setObjectName('FlowViewZoomWidget')

        self.zoom_in_button = QPushButton('+')
        self.zoom_in_button.clicked.connect(self.on_zoom_in_button_clicked)
        self.zoom_out_button = QPushButton('-')
        self.zoom_out_button.clicked.connect(self.on_zoom_out_button_clicked)

        main_horizontal_layout = QHBoxLayout()

        main_horizontal_layout.addWidget(self.zoom_out_button)
        main_horizontal_layout.addWidget(self.zoom_in_button)
        self.setLayout(main_horizontal_layout)

        self.setStyleSheet('''
QWidget#FlowViewZoomWidget {
    background: transparent; 
}
        ''')


    def on_zoom_in_button_clicked(self):
        self.flow_view.zoom_in(250)

    def on_zoom_out_button_clicked(self):
        self.flow_view.zoom_out(250)


"""Namespace for enum types and stuff"""

from enum import IntEnum


class FlowAlg(IntEnum):
    """Used for performance"""

    DATA = 1
    EXEC = 2

    @staticmethod
    def str(mode):
        if mode == FlowAlg.DATA:
            return 'data'
        elif mode == FlowAlg.EXEC:
            return 'exec'


class PortObjPos(IntEnum):
    """Used for performance"""

    INPUT = 1
    OUTPUT = 2


# registry for customizable classes
CLASSES = {
    'node base':    None,
    'data conn':    None,
    'exec conn':    None,
    'logs manager': None,
    'logger':       None,
    'vars manager': None,
    'flow':         None,
}


# class FlowVPUpdateMode(IntEnum):
#     """Used for performance"""
#
#     SYNC = 1
#     ASYNC = 2
#
#     @staticmethod
#     def stringify(mode):
#         if mode == FlowVPUpdateMode.SYNC:
#             return 'sync'
#         elif mode == FlowVPUpdateMode.ASYNC:
#             return 'async'



from qtpy.QtWidgets import QWidget, QVBoxLayout, QLineEdit, QLabel, QScrollArea
from qtpy.QtCore import Qt
from qtpy.QtGui import QFont

from ..ryvencore.InfoMsgs import InfoMsgs
from ..node_selection_widget.NodeWidget import NodeWidget


class PlaceNodeWidget(QWidget):
    """notice, that 'nodes' refers to node CLASSES here"""

    def __init__(self, flow_view, nodes):
        super(PlaceNodeWidget, self).__init__()

        self.flow_view = flow_view

        self.all_nodes = sort_nodes(nodes)  # copy, no ref

        self.nodes = []

        # 'current_nodes' are the currently selectable ones, they get recreated every time update_view() is called
        self.current_nodes = []
        self.all_current_node_widgets = []
        self.active_node_widget_index = -1
        self.active_node_widget = None

        self.reset_list()

        self.node_widget_index_counter = 0

        self.setMinimumWidth(260)
        self.setFixedHeight(450)

        self.main_layout = QVBoxLayout(self)
        self.main_layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.main_layout)

        # adding all stuff to the layout
        self.search_line_edit = QLineEdit(self)
        self.search_line_edit.setPlaceholderText('search for node...')
        self.search_line_edit.textChanged.connect(self.update_view)
        self.layout().addWidget(self.search_line_edit)


        self.list_scroll_area = QScrollArea(self)
        self.list_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.list_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.list_scroll_area.setWidgetResizable(True)

        self.list_scroll_area_widget = QWidget()
        self.list_scroll_area.setWidget(self.list_scroll_area_widget)

        self.list_layout = QVBoxLayout()
        self.list_layout.setAlignment(Qt.AlignTop)
        self.list_scroll_area_widget.setLayout(self.list_layout)

        self.layout().addWidget(self.list_scroll_area)

        self.setFixedHeight(300)


        self.update_view('')

        self.setStyleSheet(self.flow_view.session.design.node_selection_stylesheet)

        self.search_line_edit.setFocus()


    def mousePressEvent(self, event):
        QWidget.mousePressEvent(self, event)
        event.accept()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.flow_view.hide_place_node_widget()

        elif event.key() == Qt.Key_Down:
            index = self.active_node_widget_index+1 if \
                self.active_node_widget_index+1 < len(self.all_current_node_widgets) else 0
            self.set_active_node_widget_index(index)
        elif event.key() == Qt.Key_Up:
            index = self.active_node_widget_index-1 if \
                self.active_node_widget_index-1 > -1 else len(self.all_current_node_widgets)-1
            self.set_active_node_widget_index(index)

        elif event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            if len(self.all_current_node_widgets) > 0:
                self.place_node(self.active_node_widget_index)
        else:
            event.setAccepted(False)


    def wheelEvent(self, event):
        QWidget.wheelEvent(self, event)
        event.accept()


    def refocus(self):
        self.search_line_edit.setFocus()
        self.search_line_edit.selectAll()


    def update_list(self, nodes):
        self.nodes = sort_nodes(nodes)

    def reset_list(self):
        self.nodes = self.all_nodes


    def update_view(self, text=''):
        text = text.lower()
        for i in reversed(range(self.list_layout.count())):
            self.list_layout.itemAt(i).widget().setParent(None)

        self.current_nodes.clear()
        self.all_current_node_widgets.clear()

        self.node_widget_index_counter = 0

        # select valid elements from text
        # nodes
        nodes_names_dict = {}
        for n in self.nodes:
            nodes_names_dict[n] = n.title.lower()
        sorted_indices = self.get_sorted_dict_matching_search(nodes_names_dict, text)
        for n, index in sorted_indices.items():
            self.current_nodes.append(n)


        # nodes
        if len(self.current_nodes) > 0:
            # nodes_label = QLabel('Hover for description')
            # nodes_label_font = QFont('Poppins')
            # nodes_label_font.setPixelSize(15)
            # nodes_label.setStyleSheet('color: #9bbf9dff; border: none;')
            # nodes_label.setFont(nodes_label_font)
            # self.list_layout.addWidget(nodes_label)

            for n in self.current_nodes:
                node_widget = self.create_list_item_widget(n)
                self.list_layout.addWidget(node_widget)
                self.all_current_node_widgets.append(node_widget)

        if len(self.all_current_node_widgets) > 0:
            self.set_active_node_widget_index(0)

        # self.setFixedWidth(self.minimumWidth())
        # print(self.list_scroll_area_widget.width())


    def get_sorted_dict_matching_search(self, items_dict, text):
        indices_dict = {}
        for item, name in items_dict.items():  # the strings are already lowered here
            InfoMsgs.write(item, name, text)
            if name.__contains__(text):
                index = name.index(text)
                indices_dict[item] = index
        return {k: v for k, v in sorted(indices_dict.items(), key=lambda i: i[1])}


    def create_list_item_widget(self, node):
        node_widget = NodeWidget(self, node)  # , self.node_images[node])
        node_widget.custom_focused_from_inside.connect(self.node_widget_focused_from_inside)
        node_widget.setObjectName('node_widget_' + str(self.node_widget_index_counter))
        self.node_widget_index_counter += 1
        node_widget.chosen.connect(self.node_widget_chosen)
        
        return node_widget

    
    def node_widget_focused_from_inside(self):
        self.set_active_node_widget_index(self.all_current_node_widgets.index(self.sender()))

    def set_active_node_widget_index(self, index):
        self.active_node_widget_index = index
        node_widget = self.all_current_node_widgets[index]

        if self.active_node_widget:
            self.active_node_widget.set_custom_focus(False)

        node_widget.set_custom_focus(True)
        self.active_node_widget = node_widget
        self.list_scroll_area.ensureWidgetVisible(self.active_node_widget)


    def node_widget_chosen(self):  # gets called when the user clicks on a node widget with the mouse
        self.flow_view.mouse_event_taken = True  # otherwise, it will receive a mouse press event

        index = int(self.sender().objectName()[self.sender().objectName().rindex('_')+1:])
        self.place_node(index)


    def place_node(self, index):
        node_index = index
        node = self.current_nodes[node_index]
        self.flow_view.create_node__cmd(node)
        # self.flow.create_node_request.emit(node, None)

        self.flow_view.hide_place_node_widget()


def sort_nodes(nodes):
    return sorted(nodes, key=lambda x: x.title.lower())



"""The implementations for the special nodes of macros."""
from .RC import FlowAlg
from .InfoMsgs import InfoMsgs


def build_macro_classes(BaseClass):

    class MacroInputNode(BaseClass):
        """The input node of a macro script"""

        identifier = 'BUILTIN_MacroInputNode'
        title = 'input'
        type_ = 'BUILT-IN macro input node'

        def __init__(self, params):
            super().__init__(params)

            self.special_actions = {
                'add parameter': {
                    'data': {'method': self.add_macro_param, 'data': 'data'},
                    'exec': {'method': self.add_macro_param, 'data': 'exec'}
                },
                'remove parameter': {

                }
            }

        def add_macro_param(self, type_):
            self.create_output(type_, '')

            i = len(self.outputs)-1
            self.special_actions['remove parameter'][str(i)] = {
                'method': self.remove_macro_param,
                'data': i
            }

            self.script.add_parameter(type_, '')

        def remove_macro_param(self, index):
            self.delete_output(index)
            # self.rebuild_remove_actions()
            del self.special_actions['remove parameter'][str(index)]
            self.script.remove_parameter(index)

        # def rebuild_remove_actions(self):
        #     del self.special_actions['remove parameter']
        #     self.special_actions['remove parameter'] = {
        #         str(o+1): {'method': self.remove_macro_param, 'data': o} for o in range(len(self.outputs))
        #     }

        def propagate_output_values(self, values):
            self.script.output_node.hold_values()

            affected_connections = {}

            for o in range(len(self.outputs)):
                out = self.outputs[o]
                if out.type_ == 'data':
                    for c in out.connections:
                        if c.inp.val != values[o]:
                            # affected
                            affected_connections[c] = values[o]
                    self.outputs[o].val = values[o]

            for c, v in affected_connections.items():
                c.activate(v)

            self.script.output_node.release_values()


    class MacroOutputNode(BaseClass):
        """The output node of a macro script"""

        identifier = 'BUILTIN_MacroOutputNode'
        title = 'output'
        type_ = 'BUILT-IN macro output node'

        def __init__(self, params):
            super().__init__(params)

            self.special_actions = {
                'add return': {
                    'data': {'method': self.add_macro_return, 'data': 'data'},
                    'exec': {'method': self.add_macro_return, 'data': 'exec'}
                },
                'remove return': {

                }
            }

            self.held_values = {}
            self.holding_values = False
            # Basically a barrier.
            # Holding values means data input updates are not propagated
            # to the caller macro node until release_values() is called.
            # This is a nice feature for macros which allows the user
            # to split connections inside a macro (and therefore producing
            # multiple execution branches) but join all the updates again
            # in the end.
            # It is coordinated by the input node above.

        def add_macro_return(self, type_):
            self.create_input(type_)

            i = len(self.inputs)-1
            self.special_actions['remove return'][str(i)] = {
                'method': self.remove_macro_return,
                'data': i
            }

            self.script.add_return(type_, '')

        def remove_macro_return(self, index):
            self.delete_input(index)
            # self.rebuild_remove_actions()
            del self.special_actions['remove return'][str(index)]
            self.script.remove_return(index)

        # def rebuild_remove_actions(self):
        #     del self.special_actions['remove return']
        #     self.special_actions['remove return'] = {
        #         str(i+1): {'method': self.remove_macro_return, 'data': i} for i in range(len(self.inputs))
        #     }

        def update_event(self, inp=-1):
            # if len(self.script.caller_stack) == 0:
            #     return

            # caller = self.script.caller_stack[-1]

            if not self.script.caller:
                return

            caller = self.script.caller

            if inp == -1:  # requested data in exec flow
                for i in range(len(self.inputs)):
                    v = self.input(i)
                    if v != caller.outputs[i].val:  # only update if value is new
                        caller.set_output_val(i, v)

            else:   # data propagation in data flow or exec signal in exec flow
                if self.inputs[inp].type_ == 'data':
                    if self.holding_values:
                        self.held_values[self.inputs[inp]] = self.input(inp)
                else:
                    caller.exec_output(inp)

        def hold_values(self):
            self.held_values = {inp: None for inp in self.inputs}
            self.holding_values = True

        def release_values(self):
            # caller = self.script.caller_stack[-1]
            caller = self.script.caller
            for i in range(len(self.inputs)):
                inp = self.inputs[i]
                if inp.type_ == 'data':
                    v = inp.val
                    if v != caller.outputs[i].val:  # only update if value is new
                        caller.set_output_val(i, inp.val)


    class MacroScriptNode(BaseClass):
        """The macro node of a macro script"""

        type_ = 'BUILT-IN macro node'

        def __init__(self, params):
            super().__init__(params)

            self.instances.append(self)  # defined statically in the custom macro node class

            if not self.init_config:
                # catch up on params and returns
                for p in self.macro_script.parameters:
                    self.create_input(p['type'], p['label'])
                for r in self.macro_script.returns:
                    self.create_output(r['type'], r['label'])

        def update_event(self, inp=-1):

            self.macro_script.caller = self

            if inp != -1:

                # self.macro_script.caller_stack.append(self)

                if self.flow.alg_mode == FlowAlg.DATA:
                    values = [inp.val for inp in self.inputs]
                    InfoMsgs.write('sending propagation values to input node', values)
                    self.macro_script.input_node.propagate_output_values(values)

                    # if somebody wants to use exec inputs in a data flow for some reason
                    if self.inputs[inp].type_ == 'exec':
                        self.macro_script.input_node.exec_output(inp)
                else:
                    # manually set output values of input node
                    for i in range(len(self.inputs)):
                        if self.inputs[i].type_ == 'data':
                            self.macro_script.input_node.outputs[i].val = self.inputs[i].val
                    self.macro_script.input_node.exec_output(inp)

                # self.macro_script.caller_stack.pop()

            else:
                # self.macro_script.caller_stack.append(self)
                self.macro_script.output_node.update()
                # self.macro_script.caller_stack.pop()

            self.macro_script.caller = None

        def get_state(self):
            data = {'title': self.macro_script.title}
            return data

        def set_state(self, data: dict):
            # find parent macro script by unique title
            for ms in self.session.macro_scripts:
                if ms.title == data['title']:
                    self.macro_script = ms
                    break


    return MacroInputNode, MacroOutputNode, MacroScriptNode



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QWidget, QMessageBox, QVBoxLayout, QLineEdit, QHBoxLayout, QPushButton, QScrollArea

from .ScriptsList_ScriptWidget import ScriptsList_ScriptWidget


class ScriptsListWidget(QWidget):
    """Convenience class for a QWidget to easily manage the scripts of a session."""

    # TODO: scroll area

    def __init__(self, session, parent=None):
        super(ScriptsListWidget, self).__init__(parent=parent)

        self.session = session
        self.session.flow_view_created.connect(self.add_new_script)
        self.session.script_deleted.connect(self.recreate_list)

        self.list_widgets = []
        self.ignore_name_line_edit_signal = False  # because disabling causes firing twice otherwise

        self.setup_UI()


    def setup_UI(self):
        main_layout = QVBoxLayout()

        self.list_layout = QVBoxLayout()
        self.list_layout.setAlignment(Qt.AlignTop)

        # list scroll area

        self.list_scroll_area = QScrollArea()
        self.list_scroll_area.setWidgetResizable(True)
        self.list_scroll_area.setContentsMargins(0, 0, 0, 0)

        w = QWidget()
        w.setContentsMargins(0, 0, 0, 0)
        w.setLayout(self.list_layout)

        self.list_scroll_area.setWidget(w)

        main_layout.addWidget(self.list_scroll_area)

        # ------------------

        # controls

        self.new_script_title_lineedit = QLineEdit()
        self.new_script_title_lineedit.setPlaceholderText('new script\'s title')
        self.new_script_title_lineedit.returnPressed.connect(self.new_script_LE_return_pressed)

        main_layout.addWidget(self.new_script_title_lineedit)

        buttons_layout = QHBoxLayout()

        self.create_macro_button = QPushButton('macro')
        self.create_macro_button.clicked.connect(self.on_create_macro_clicked)
        buttons_layout.addWidget(self.create_macro_button)

        self.create_script_button = QPushButton('script')
        self.create_script_button.clicked.connect(self.on_create_script_clicked)
        buttons_layout.addWidget(self.create_script_button)

        main_layout.addLayout(buttons_layout)

        # ------------------

        self.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_layout)

        self.recreate_list()


    def recreate_list(self):
        for w in self.list_widgets:
            w.hide()
            del w

        self.list_widgets.clear()

        for s in self.session.scripts:  # all_scripts():
            new_widget = ScriptsList_ScriptWidget(self, self.session, s)
            self.list_widgets.append(new_widget)

        self.rebuild_list()


    def rebuild_list(self):
        # for i in range(self.layout().count()):
        #     self.list_layout.removeItem(self.layout().itemAt(0))

        for w in self.list_widgets:
            self.list_layout.addWidget(w)


    def new_script_LE_return_pressed(self):
        self.create_script()  # create normal scripts by default

    def on_create_macro_clicked(self):
        self.create_macro_script()

    def on_create_script_clicked(self):
        self.create_script()

    def create_macro_script(self):
        title = self.new_script_title_lineedit.text()

        if self.session.script_title_valid(title):
            self.session.create_macro(title=title)

    def create_script(self):
        title = self.new_script_title_lineedit.text()

        if self.session.script_title_valid(title):
            self.session.create_script(title=title)


    def add_new_script(self, script):
        self.recreate_list()


    def del_script(self, script, script_widget):
        msg_box = QMessageBox(QMessageBox.Warning, 'sure about deleting script?',
                              'You are about to delete a script. This cannot be undone, all content will be lost. '
                              'Do you want to continue?', QMessageBox.Cancel | QMessageBox.Yes, self)
        msg_box.setDefaultButton(QMessageBox.Cancel)
        ret = msg_box.exec_()
        if ret != QMessageBox.Yes:
            return

        self.list_widgets.remove(script_widget)
        script_widget.setParent(None)
        self.session.delete_script(script)
        # self.recreate_list()



from qtpy.QtCore import Signal
from qtpy.QtGui import QFontMetrics, QFont
from qtpy.QtWidgets import QSpinBox, QLineEdit, QSlider, QCheckBox, QComboBox


from .WidgetBaseClasses import IWB
from .retain import M


class RCIW_BUILTIN_SpinBox(QSpinBox, IWB):
    """Convenience class for using a spin box widget as data input for nodes."""

    trigger_update = Signal(int)

    def __init__(self, params):
        QSpinBox.__init__(self)
        IWB.__init__(self, params)

        self.trigger_update.connect(self.node.update)

        self.port_local_pos = None

        self.setFixedWidth(50)
        self.setFixedHeight(25)
        self.setMaximum(1000000)
        self.editingFinished.connect(self.editing_finished)

    def editing_finished(self):
        self.trigger_update.emit(self.node.inputs.index(self.input))

    def remove_event(self):
        pass

    def get_val(self):
        return self.value()

    def get_state(self) -> dict:
        return {'val': self.value()}

    def set_state(self, data):
        if type(data) == dict:
            self.setValue(data['val'])
        else:  # backwards compatibility
            self.setValue(data)


class RCIW_BUILTIN_LineEdit(QLineEdit, IWB):
    """Convenience class for using input fields (i.e. QLineEdits) as widgets for data inputs of nodes."""

    trigger_update = Signal(object)

    def __init__(self, params, size='medium', resize=False):
        IWB.__init__(self, params)
        QLineEdit.__init__(self)

        self.trigger_update.connect(self.input.update)

        self.port_local_pos = None
        self.resizing = resize

        if size == 'small':
            self.base_width = 30
        elif size == 'medium':
            self.base_width = 70
        elif size == 'large':
            self.base_width = 150

        self.setFixedWidth(self.base_width)

        # self.setFixedHeight(25)
        self.setPlaceholderText('')

        f = self.font()
        f.setPointSize(10)
        self.setFont(f)
        self.textChanged.connect(self.text_changed)
        self.editingFinished.connect(self.editing_finished)

    def text_changed(self, new_text):
        if self.resizing:
            fm = QFontMetrics(self.font())
            text_width = fm.width(new_text)
            new_width = text_width+15
            self.setFixedWidth(new_width if new_width > self.base_width else self.base_width)

            self.node.update_shape()
            # self.parent_node_instance.rebuild_ui()  # see rebuild_ui() for explanation

    def editing_finished(self):
        # self.node.update(self.node.inputs.index(self.input))
        self.trigger_update.emit(self.get_val())

    def remove_event(self):
        pass

    def get_val(self):
        try:
            val = eval(self.text())
        except Exception as e:
            # type(eval(json.dumps(self.text()))) could be 'dict' <- need that for typing in dicts later if I want to
            val = self.text()
        return val

    def get_state(self) -> dict:
        return {'text': self.text()}

    def set_state(self, data):
        if type(data) == str:  # backwards compatibility
            self.setText(data)
        elif type(data) == dict:
            self.setText(data['text'])

    def val_update_event(self, val):
        self.setText(str(val))


class RCIW_BUILTIN_LineEdit_small(RCIW_BUILTIN_LineEdit):
    pass






class DType_IW_Base(IWB):
    def __str__(self):
        return self.__class__.__name__




class Data_IW(QLineEdit, DType_IW_Base):  # virtual

    base_width = None  # specified by subclasses

    def __init__(self, params):
        QLineEdit.__init__(self)
        DType_IW_Base.__init__(self, params)

        dtype = self.input.dtype

        self.setFont(QFont('source code pro', 10))
        self.val_update_event(dtype.default)
        if dtype.size == 's':
            self.base_width = 30
        elif dtype.size == 'm':
            self.base_width = 70
        elif dtype.size == 'l':
            self.base_width = 150
        self.max_width = self.base_width * 3
        self.setFixedWidth(self.base_width)
        self.fm = QFontMetrics(self.font())

        self.setToolTip(dtype.doc)
        self.textChanged.connect(self.text_changed)
        self.editingFinished.connect(self.editing_finished)

    def text_changed(self, new_text):
        text_width = self.fm.width(new_text)
        new_width = text_width+15  # add some buffer
        if new_width < self.max_width:
            self.setFixedWidth(new_width if new_width > self.base_width else self.base_width)
        else:
            self.setFixedWidth(self.max_width)
        self.node.update_shape()

    def editing_finished(self):
        self.update_node_input(self.get_val())

    def get_val(self):
        try:
            return eval(self.text())
        except Exception as e:
            return self.text()

    def val_update_event(self, val):
        try:
            self.setText(str(val))
        except Exception as e:
            pass

    def get_state(self) -> dict:
        return {'text': self.text()}

    def set_state(self, data: dict):
        self.setText(data['text'])


# custom sized classes for qss access:

class Data_IW_S(Data_IW):
    base_width = 30


class Data_IW_M(Data_IW):
    base_width = 70


class Data_IW_L(Data_IW):
    base_width = 150


# -----------------------------------


class String_IW(QLineEdit, DType_IW_Base):  # virtual

    width = None  # specified by subclasses

    def __init__(self, params):
        QLineEdit.__init__(self)
        DType_IW_Base.__init__(self, params)

        dtype = self.input.dtype

        self.setFont(QFont('source code pro', 10))
        self.setText(dtype.default)
        self.setFixedWidth(self.width)
        self.setToolTip(dtype.doc)

        self.block = False  # ignore updates from val_update_event
        self.editingFinished.connect(self.editing_finished)

    def editing_finished(self):
        if not self.block:
            self.update_node_input(self.get_val())

    def get_val(self):
        return self.text()

    def val_update_event(self, val):
        self.block = True
        self.setText(str(val))
        self.block = False

    def get_state(self) -> dict:
        return {'text': self.text()}

    def set_state(self, data: dict):
        self.setText(data['text'])


# custom sized classes for qss access:

class String_IW_S(String_IW):
    width = 30


class String_IW_M(String_IW):
    width = 70


class String_IW_L(String_IW):
    width = 150


# -----------------------------------


class Integer_IW(QSpinBox, DType_IW_Base):
    def __init__(self, params):
        QSpinBox.__init__(self)
        DType_IW_Base.__init__(self, params)

        dtype = self.input.dtype

        if dtype.bounds:
            self.setRange(dtype.bounds[0], dtype.bounds[1])
        self.setValue(dtype.default)
        self.setToolTip(dtype.doc)

        self.block = False  # ignore updates from val_update_event
        self.valueChanged.connect(self.val_changed)

    def val_changed(self, val):
        if not self.block:
            self.update_node_input(val)

    def get_val(self):
        return self.value()

    def val_update_event(self, val):
        self.block = True
        try:
            self.setValue(val)
        finally:
            self.block = False

    def get_state(self) -> dict:
        return {'val': self.value()}

    def set_state(self, data: dict):
        self.setValue(data['val'])


class Float_IW(QLineEdit, DType_IW_Base):
    def __init__(self, params):
        QLineEdit.__init__(self)
        DType_IW_Base.__init__(self, params)

        dtype = self.input.dtype

        self.setFont(QFont('source code pro', 10))
        fm = QFontMetrics(self.font())
        self.setMaximumWidth(fm.width(' ')*dtype.decimals+1)
        self.setText(str(dtype.default))
        self.setToolTip(dtype.doc)

        self.block = False  # ignore updates from val_update_event
        self.textChanged.connect(self.text_changed)

    def text_changed(self):
        if not self.block:
            self.update_node_input(self.get_val())

    def get_val(self):
        return float(self.text())

    def val_update_event(self, val):
        self.block = True
        try:
            self.setText(str(val))
        finally:
            self.block = False

    def get_state(self) -> dict:
        return {'text': self.text()}

    def set_state(self, data: dict):
        self.setText(data['text'])


class Boolean_IW(QCheckBox, DType_IW_Base):
    def __init__(self, params):
        QCheckBox.__init__(self)
        DType_IW_Base.__init__(self, params)

        dtype = self.input.dtype

        self.setChecked(dtype.default)

        self.setToolTip(dtype.doc)

        self.block = False  # ignore updates from val_update_event
        self.stateChanged.connect(self.state_changed)

    def state_changed(self, state):
        if not self.block:
            self.update_node_input(self.get_val())

    def get_val(self):
        return self.isChecked()

    def val_update_event(self, val):
        self.block = True
        try:
            self.setChecked(bool(val))
        finally:
            self.block = False

    def get_state(self) -> dict:
        return {'checked': self.isChecked()}

    def set_state(self, data: dict):
        self.setChecked(data['checked'])


class Choice_IW(QComboBox, DType_IW_Base):
    def __init__(self, params):
        QComboBox.__init__(self)
        DType_IW_Base.__init__(self, params)

        dtype = self.input.dtype

        self.addItems(dtype.items)
        self.setCurrentText(dtype.default)
        self.setToolTip(dtype.doc)

        self.block = False  # ignore updates from val_update_event
        self.currentTextChanged.connect(self.text_changed)

    def text_changed(self):
        if not self.block:
            self.update_node_input(self.get_val())

    def get_val(self):
        return self.currentText()

    def val_update_event(self, val):
        self.block = True
        try:
            self.setCurrentText(val)
        finally:
            self.block = False

    def get_state(self) -> dict:
        return {
            'items': [self.itemText(i) for i in range(self.count())],
            'active': self.currentText(),
        }

    def set_state(self, data: dict):
        self.clear()
        self.addItems(data['items'])
        self.setCurrentText(data['active'])



import json

from qtpy.QtCore import QObject, Signal
from qtpy.QtGui import QFontDatabase

from .FlowTheme import FlowTheme_Toy, FlowTheme_DarkTron, FlowTheme_Ghost, FlowTheme_Blender, \
    FlowTheme_Simple, FlowTheme_Ueli, FlowTheme_PureDark, FlowTheme, FlowTheme_PureLight, \
    FlowTheme_Colorful, FlowTheme_ColorfulLight
from .GlobalAttributes import Location


class Design(QObject):
    """Design serves as a container for the stylesheet and flow themes, and sends signals to notify GUI elements
    on change of the flow theme. A configuration for the flow themes can be loaded from a json file."""

    global_stylesheet = ''

    flow_theme_changed = Signal(str)
    performance_mode_changed = Signal(str)

    def __init__(self):
        super().__init__()

        self.flow_themes = []
        self.flow_theme: FlowTheme = None
        self.default_flow_size = None
        self.performance_mode: str = None
        self.node_item_shadows_enabled: bool = None
        self.animations_enabled: bool = None
        self.node_selection_stylesheet: str = None

        # load standard default values
        self.register_flow_themes()
        self._default_flow_theme = self.flow_themes[-1]
        self.set_performance_mode('pretty')
        self.set_animations_enabled(True)
        self.default_flow_size = [1000, 700]
        self.set_flow_theme(self._default_flow_theme)

    @staticmethod
    def register_fonts():
        db = QFontDatabase()
        db.addApplicationFont(
            Location.PACKAGE_PATH + '/resources/fonts/poppins/Poppins-Medium.ttf'
        )
        db.addApplicationFont(
            Location.PACKAGE_PATH + '/resources/fonts/source_code_pro/SourceCodePro-Regular.ttf'
        )
        db.addApplicationFont(
            Location.PACKAGE_PATH + '/resources/fonts/asap/Asap-Regular.ttf'
        )

    def register_flow_themes(self):
        self.flow_themes = [
            FlowTheme_Toy(),
            FlowTheme_DarkTron(),
            FlowTheme_Ghost(),
            FlowTheme_Blender(),
            FlowTheme_Simple(),
            FlowTheme_Ueli(),
            FlowTheme_PureDark(),
            FlowTheme_Colorful(),
            FlowTheme_PureLight(),
            FlowTheme_ColorfulLight()
        ]

    def load_from_config(self, filepath: str):
        """Loads design configs from a config json file"""

        f = open(filepath, 'r')
        data = f.read()
        f.close()

        IMPORT_DATA = json.loads(data)

        if 'flow themes' in IMPORT_DATA:
            # load flow theme configs
            FTID = IMPORT_DATA['flow themes']
            for flow_theme in self.flow_themes:
                flow_theme.load(FTID)

        if 'init flow theme' in IMPORT_DATA:
            self._default_flow_theme = self.flow_theme_by_name(IMPORT_DATA.get('init flow theme'))
            self.set_flow_theme(self._default_flow_theme)

        if 'init performance mode' in IMPORT_DATA:
            self.set_performance_mode(IMPORT_DATA['init performance mode'])

        if 'init animations enabled' in IMPORT_DATA:
            self.set_animations_enabled(IMPORT_DATA['init animations enabled'])

        if 'default flow size' in IMPORT_DATA:
            self.default_flow_size = IMPORT_DATA['default flow size']

    def available_flow_themes(self) -> dict:
        return {theme.name: theme for theme in self.flow_themes}

    def flow_theme_by_name(self, name: str) -> FlowTheme:
        for theme in self.flow_themes:
            if theme.name == name:
                return theme
        return None

    def set_flow_theme(self, theme: FlowTheme = None, name: str = ''):
        """You can either specify the theme by name, or directly provide a FlowTheme object"""
        if theme:
            self.flow_theme = theme
        elif name and name != '':
            self.flow_theme = self.flow_theme_by_name(name)
        else:
            return

        self.node_selection_stylesheet = self.flow_theme.build_node_selection_stylesheet()

        self.flow_theme_changed.emit(self.flow_theme.name)


    def set_performance_mode(self, new_mode: str):
        self.performance_mode = new_mode
        if new_mode == 'fast':
            self.node_item_shadows_enabled = False
        else:
            self.node_item_shadows_enabled = True

        self.performance_mode_changed.emit(self.performance_mode)

    def set_animations_enabled(self, b: bool):
        self.animations_enabled = b

    def set_node_item_shadows(self, b: bool):
        self.node_item_shadows_enabled = b





# default_node_selection_stylesheet = '''
# '''



from ..Base import Base

from ..RC import CLASSES

# from logging import Logger
from .Logger import Logger


class LogsManager(Base):
    """Manages all logs/loggers that belong to the script."""

    def __init__(self, script, create_default_logs=True):
        Base.__init__(self)

        self.script = script
        self.loggers: [Logger] = []
        self.default_loggers = {
            'global': None,
            'errors': None,
        }

        if create_default_logs:
            self.create_default_loggers()

    def create_default_loggers(self):
        for name in self.default_loggers.keys():
            self.default_loggers[name] = self.new_logger(title=name.title())

    def new_logger(self, title: str) -> Logger:
        logger = CLASSES['logger'](name=title)
        self.loggers.append(logger)
        return logger



from qtpy.QtCore import QObject, Signal

from .ryvencore import Node as RC_Node
from .GlobalAttributes import Location
from .ryvencore.dtypes import DType


class Node(RC_Node, QObject):
    """Base class nodes in ryvencore-qt"""

    description_html: str = None
    main_widget_class: list = None
    main_widget_pos: str = 'below ports'
    input_widget_classes: dict = {}
    style: str = 'normal'
    color: str = '#c69a15'
    icon: str = None

    # SIGNALS
    updated = Signal()
    input_added = Signal(object, object)
    output_added = Signal(object, object)
    input_removed = Signal(object)
    output_removed = Signal(object)
    update_shape_triggered = Signal()
    hide_unconnected_ports_triggered = Signal()
    show_unconnected_ports_triggered = Signal()

    def __init__(self, params):
        QObject.__init__(self)
        RC_Node.__init__(self, params)

        self.default_actions = self.init_default_actions()
        self.special_actions = {}
        self.display_title = self.title

        self.item = None  # set by the flow view

    @staticmethod
    def complete_default_node_classes():
        func_node_icon = Location.PACKAGE_PATH + '/resources/pics/macro_node_icon.png'
        from .ryvencore.MacroScript import MacroScript
        MacroScript.MacroNode.icon = func_node_icon

    """actions"""

    def init_default_actions(self) -> dict:
        actions = {
            'update shape': {'method': self.update_shape},
            'hide unconnected ports': {'method': self.hide_unconnected_ports},
            'change title': {'method': self.change_title}
        }
        return actions


    def set_special_actions_data(self, actions_data):
        actions = {}
        for key in actions_data:
            if type(actions_data[key]) != dict:
                if key == 'method':
                    try:
                        actions['method'] = getattr(self, actions_data[key])
                    except AttributeError:  # outdated method referenced
                        pass
                elif key == 'data':
                    actions['data'] = actions_data[key]
            else:
                actions[key] = self.set_special_actions_data(actions_data[key])
        return actions


    def get_special_actions_data(self, actions):
        cleaned_actions = actions.copy()
        for key in cleaned_actions:
            v = cleaned_actions[key]
            # if type(v) == M:  # callable(v):
            #     cleaned_actions[key] = v.method_name
            if callable(v):
                cleaned_actions[key] = v.__name__
            elif type(v) == dict:
                cleaned_actions[key] = self.get_special_actions_data(v)
            else:
                cleaned_actions[key] = v
        return cleaned_actions


    def get_extended_default_actions(self):
        actions_dict = self.default_actions.copy()
        for index in range(len(self.inputs)):
            inp = self.inputs[index]
            if inp.type_ == 'exec':
                actions_dict['exec input '+str(index)] = {'method': self.action_exec_input,
                                                          'data': {'input index': index}}
        return actions_dict


    def action_exec_input(self, data):
        self.update(data['input index'])

    """
    specifications of ryvencore.Node's default behavior:
    """

    # @override
    def update(self, inp=-1):
        RC_Node.update(self, inp)
        self.updated.emit()

    # @override
    def create_input(self, label: str = '', type_: str = 'data', add_config={}, insert: int = None):
        RC_Node.create_input(self, label=label, type_=type_, add_config=add_config, insert=insert)

        if insert is not None:

            if insert < 0:
                index = insert-1
            else:
                index = insert

            self.input_added.emit(self.inputs[index], insert)

        else:
            self.input_added.emit(self.inputs[-1], None)

    # @override
    def create_input_dt(self, label: str, dtype: DType, add_config={}, insert: int = None):
        RC_Node.create_input_dt(self, label=label, dtype=dtype, add_config=add_config, insert=insert)

        if insert is not None:

            if insert < 0:
                index = insert-1
            else:
                index = insert

            self.input_added.emit(self.inputs[index], insert)
        else:
            self.input_added.emit(self.inputs[-1], None)

    # @override
    def create_output(self, label: str = '', type_: str = 'data', insert: int = None):
        RC_Node.create_output(self, label=label, type_=type_, insert=insert)

        if insert is not None:

            if insert < 0:
                index = insert-1
            else:
                index = insert

            self.output_added.emit(self.outputs[index], insert)
        else:
            self.output_added.emit(self.outputs[-1], None)

    # @override
    def delete_input(self, index):
        inp = self.inputs[index]
        RC_Node.delete_input(self, index=index)
        self.input_removed.emit(inp)

    # @override
    def delete_output(self, index):
        out = self.outputs[index]
        RC_Node.delete_output(self, index=index)
        self.output_removed.emit(out)

    # @override
    def custom_config_data(self) -> dict:
        return {
            'special actions': self.get_special_actions_data(self.special_actions),
            'display title': self.display_title,
        }

    # @override
    def load_custom_config(self, data: dict):
        self.special_actions = self.set_special_actions_data(data['special actions'])
        self.display_title = data['display title']

    # @override
    def input(self, index: int):
        if len(self.inputs[index].connections) == 0 and self.item:
            iw = self.input_widget(index)
            return iw.get_val() if iw else None
        else:
            return RC_Node.input(self, index)

    # @override
    def prepare_removal(self):  # gets also subclassed by user
        if self.main_widget():
            self.main_widget().remove_event()
        RC_Node.prepare_removal(self)


    """
    additional stuff for GUI access:
    [everything below is pure ryvencore-qt API and, to ensure ryvencore compatibility, 
    should not be used unchecked in nodes]
    """


    def set_display_title(self, t: str):
        self.display_title = t
        self.update_shape()


    def flow_view(self):
        """Returns the registered FlowView of the parent script, but None if the view isn't existent
        which can happen if you call this early"""

        return self.session.flow_views[self.flow.script] if self.flow.script in self.session.flow_views else None


    def main_widget(self):
        """Returns the main_widget object, or None if the item doesn't exist (yet)"""
        if self.item:
            return self.item.main_widget
        else:
            return None


    def has_main_widget(self):
        return self.main_widget() is not None


    def input_widget(self, index: int):
        """Returns a reference to the widget of the corresponding input"""

        return self.port_item(self.inputs[index]).widget


    def session_stylesheet(self) -> str:
        """Returns the registered stylesheet of the session"""

        return self.session.design.global_stylesheet


    def port_item(self, port_obj):
        """Returns the port item object associated with a given port object"""

        if port_obj in self.inputs:
            return self.item.inputs[self.inputs.index(port_obj)]
        elif port_obj in self.outputs:
            return self.item.outputs[self.outputs.index(port_obj)]
        return None


    def update_shape(self):
        """Causes recompilation of the whole shape of the GUI item."""

        self.update_shape_triggered.emit()


    def hide_unconnected_ports(self):
        """Causes the GUI item to hide all unconnected ports"""

        del self.default_actions['hide unconnected ports']
        self.default_actions['show unconnected ports'] = {'method': self.show_unconnected_ports}
        self.hide_unconnected_ports_triggered.emit()


    def show_unconnected_ports(self):
        """Causes the GUI item to show all unconnected ports that have been hidden previously"""

        del self.default_actions['show unconnected ports']
        self.default_actions['hide unconnected ports'] = {'method': self.hide_unconnected_ports}
        self.show_unconnected_ports_triggered.emit()


    def change_title(self):
        from qtpy.QtWidgets import QDialog, QVBoxLayout, QLineEdit

        class ChangeTitleDialog(QDialog):
            def __init__(self, title):
                super().__init__()
                self.new_title = None
                self.setLayout(QVBoxLayout())
                self.line_edit = QLineEdit(title)
                self.layout().addWidget(self.line_edit)
                self.line_edit.returnPressed.connect(self.return_pressed)

            def return_pressed(self):
                self.new_title = self.line_edit.text()
                self.accept()

        d = ChangeTitleDialog(self.display_title)
        d.exec_()
        if d.new_title:
            self.set_display_title(d.new_title)



from qtpy.QtWidgets import QWidget, QLabel, QGridLayout, QHBoxLayout, QSpacerItem, QSizePolicy, QStyleOption, QStyle
from qtpy.QtGui import QFont, QPainter, QColor
from qtpy.QtCore import Signal


class NodeWidget(QWidget):

    chosen = Signal()
    custom_focused_from_inside = Signal()

    def __init__(self, parent, node):
        super(NodeWidget, self).__init__(parent)

        self.custom_focused = False
        self.node = node
#         self.custom_focused_stylesheet = '''
# NodeWidget {
#     border: 2px solid #49aeed;
#     background-color: #246187;
# }
#         '''
#         self.custom_unfocused_stylesheet = '''
# NodeWidget {
#     border: 2px solid #3d3d3d;
# }
#         '''

        # UI
        main_layout = QGridLayout()

        name_label = QLabel(node.title)
        name_label.setFont(QFont('Poppins', 12))

        type_layout = QHBoxLayout()

        type_label = QLabel(node.type_)
        type_label.setFont(QFont('Arial', 8, italic=True))
        # type_label.setStyleSheet('color: white;')

        package_name_layout_spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        type_layout.addItem(package_name_layout_spacer)
        type_layout.addWidget(type_label)

        main_layout.addWidget(name_label, 0, 0)
        main_layout.addLayout(type_layout, 1, 0)


        # # ------------------------------------------
        # img_label = QLabel()
        # img_label.setStyleSheet('padding: 20px;')
        # pix = QPixmap(self.node_image)
        # img_label.setPixmap(pix)
        # main_layout.addWidget(img_label, 2, 0)
        # # ------------------------------------------


        self.setLayout(main_layout)


        main_layout.setVerticalSpacing(0)
        main_layout.setSpacing(0)
        self.setContentsMargins(-6, -6, -6, -6)
        main_layout.setContentsMargins(-6, -6, -6, -6)
        self.setToolTip(node.__doc__)
        self.update_stylesheet()
        # self.setMinimumWidth(70)



    def mousePressEvent(self, event):
        self.custom_focused_from_inside.emit()

    # def mouseMoveEvent(self, event):
    #     self.set_custom_focus(True)

    def mouseReleaseEvent(self, event):
        if self.geometry().contains(self.mapToParent(event.pos())):
            self.chosen.emit()

    def set_custom_focus(self, new_focus):
        self.custom_focused = new_focus
        self.update_stylesheet()
        # if self.custom_focused:
        #     self.setFocus()

        # if new_focus:
        #     self.setStyleSheet(self.custom_focused_stylesheet + '\n' + self.contents_stylesheet)
        # else:
        #     self.setStyleSheet(self.custom_unfocused_stylesheet + '\n' + self.contents_stylesheet)

    def update_stylesheet(self):
        new_style_sheet = f'''
NodeWidget {{
    border: 2px solid rgba({(
        f'{QColor(self.node.color).red()},{QColor(self.node.color).green()},{QColor(self.node.color).blue()},150'
    )});
    border-radius: 4px;
    {(
        f'background-color: rgba({QColor(self.node.color).red()},{QColor(self.node.color).green()},{QColor(self.node.color).blue()},50);'
    ) if self.custom_focused else ''}
}}   
QLabel {{
    color: {self.node.color};
    background: transparent;
}}
        '''

        self.setStyleSheet(new_style_sheet)

    def paintEvent(self, event):  # just to enable stylesheets
        o = QStyleOption()
        o.initFrom(self)
        p = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, o, p, self)


import time
import json

from qtpy.QtCore import Qt, QPointF, QPoint, QRectF, QSizeF, Signal, QTimer
from qtpy.QtGui import QPainter, QPen, QColor, QKeySequence, QTabletEvent, QImage, QGuiApplication, QFont
from qtpy.QtWidgets import QGraphicsView, QGraphicsScene, QShortcut, QMenu, QGraphicsItem, QUndoStack

from .ryvencore.Flow import Flow
from .FlowCommands import MoveComponents_Command, PlaceNode_Command, \
    PlaceDrawing_Command, RemoveComponents_Command, ConnectPorts_Command, Paste_Command, FlowUndoCommand
from .FlowViewProxyWidget import FlowViewProxyWidget
from .FlowViewStylusModesWidget import FlowViewStylusModesWidget
from .FlowSessionThreadInterface import FlowSessionThreadInterface
from .FlowViewZoomWidget import FlowViewZoomWidget
from .ryvencore.Node import Node
from .ryvencore.NodePort import NodePort
from .node_selection_widget.PlaceNodeWidget import PlaceNodeWidget
from .NodeItem import NodeItem
from .PortItem import PortItemPin, PortItem
from .ryvencore.Connection import Connection, DataConnection
from .ConnectionItem import default_cubic_connection_path, ConnectionItem
from .DrawingObject import DrawingObject
from .ryvencore.InfoMsgs import InfoMsgs
from .ryvencore.RC import PortObjPos, CLASSES


# from .ryvencore.RC import FlowVPUpdateMode as VPUpdateMode


class FlowView(QGraphicsView):
    """Manages the GUI of flows"""

    nodes_selection_changed = Signal(list)
    node_placed = Signal(Node)

    create_node_request = Signal(object, dict)
    remove_node_request = Signal(Node)

    check_connection_validity_request = Signal(NodePort, NodePort, bool)
    connect_request = Signal(NodePort, NodePort)

    get_nodes_config_request = Signal(list)
    get_connections_config_request = Signal(list)
    get_flow_config_data_request = Signal()

    viewport_update_mode_changed = Signal(str)

    def __init__(self, session, script, flow, config=None, flow_size: list = None, parent=None):
        super(FlowView, self).__init__(parent=parent)

        # UNDO/REDO
        self._undo_stack = QUndoStack(self)
        self._undo_action = self._undo_stack.createUndoAction(self, 'undo')
        self._undo_action.setShortcuts(QKeySequence.Undo)
        self._redo_action = self._undo_stack.createRedoAction(self, 'redo')
        self._redo_action.setShortcuts(QKeySequence.Redo)

        # SHORTCUTS
        self._init_shortcuts()

        # GENERAL ATTRIBUTES
        self.session = session
        self.script = script
        self.flow: Flow = flow
        self.node_items: dict = {}  # {Node: NodeItem}
        self.node_items__cache: dict = {}
        self.connection_items: dict = {}  # {Connection: ConnectionItem}
        self.connection_items__cache: dict = {}

        # PRIVATE FIELDS
        self._tmp_data = None
        # self._temp_config_data = None
        self._selected_pin: PortItemPin = None
        self._dragging_connection = False
        self._temp_connection_ports = None
        self.mouse_event_taken = False  # for stylus - see tablet event
        self._showing_framerate = False
        self._last_mouse_move_pos: QPointF = None
        self._node_place_pos = QPointF()
        self._left_mouse_pressed_in_flow = False
        self._right_mouse_pressed_in_flow = False
        self._mouse_press_pos: QPointF = None
        self._auto_connection_pin = None  # stores the gate that we may try to auto connect to a newly placed NI
        self._panning = False
        self._pan_last_x = None
        self._pan_last_y = None
        self._current_scale = 1
        self._total_scale_div = 1

        # CONNECTIONS TO FLOW
        self.create_node_request.connect(self.flow.create_node)
        self.remove_node_request.connect(self.flow.remove_node)
        self.node_placed.connect(self.flow.node_view_placed)
        self.check_connection_validity_request.connect(self.flow.check_connection_validity)
        self.get_nodes_config_request.connect(self.flow.generate_nodes_config)
        self.get_connections_config_request.connect(self.flow.generate_connections_config)
        self.get_flow_config_data_request.connect(self.flow.generate_config_data)

        # CONNECTIONS FROM FLOW
        self.flow.node_added.connect(self.add_node)
        self.flow.node_removed.connect(self.remove_node)
        self.flow.connection_added.connect(self.add_connection)
        self.flow.connection_removed.connect(self.remove_connection)
        self.flow.connection_request_valid.connect(self.connection_request_valid)
        # self.flow.nodes_config_generated.connect(self._abstract_nodes_config_generated)
        # self.flow.connections_config_generated.connect(self._abstract_connections_config_generated)

        # SESSION THREAD
        # if self.session.threaded:
        self.thread_interface = FlowSessionThreadInterface()
        self.thread_interface.moveToThread(self.session.thread())

        # # SETTINGS
        # self.vp_update_mode: VPUpdateMode = VPUpdateMode.SYNC

        # CREATE UI
        scene = QGraphicsScene(self)
        scene.setItemIndexMethod(QGraphicsScene.NoIndex)
        if flow_size is None:
            scene.setSceneRect(0, 0, 10 * self.width(), 10 * self.height())
        else:
            scene.setSceneRect(0, 0, flow_size[0], flow_size[1])

        self.setScene(scene)
        self.setCacheMode(QGraphicsView.CacheBackground)
        self.setViewportUpdateMode(QGraphicsView.BoundingRectViewportUpdate)
        self.setRenderHint(QPainter.Antialiasing)
        self.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)
        self.setDragMode(QGraphicsView.RubberBandDrag)
        scene.selectionChanged.connect(self._scene_selection_changed)
        self.setAcceptDrops(True)

        self.centerOn(QPointF(self.viewport().width() / 2, self.viewport().height() / 2))

        # PLACE NODE WIDGET
        self._place_node_widget_proxy = FlowViewProxyWidget(self)
        self._place_node_widget_proxy.setZValue(1000)
        self._place_node_widget = PlaceNodeWidget(self, self.session.nodes)
        self._place_node_widget_proxy.setWidget(self._place_node_widget)
        self.scene().addItem(self._place_node_widget_proxy)
        self.hide_place_node_widget()

        # ZOOM WIDGET
        # self._zoom_proxy = FlowViewProxyWidget(self)
        # self._zoom_proxy.setFlag(QGraphicsItem.ItemIgnoresTransformations, True)
        # self._zoom_proxy.setZValue(1001)
        # self._zoom_widget = FlowViewZoomWidget(self)
        # self._zoom_proxy.setWidget(self._zoom_widget)
        # self.scene().addItem(self._zoom_proxy)
        # self.set_zoom_proxy_pos()

        # STYLUS
        self.stylus_mode = ''
        self._current_drawing = None
        self._drawing = False
        self.drawings = []
        self._stylus_modes_proxy = FlowViewProxyWidget(self)
        self._stylus_modes_proxy.setFlag(QGraphicsItem.ItemIgnoresTransformations, True)
        self._stylus_modes_proxy.setZValue(1001)
        self._stylus_modes_widget = FlowViewStylusModesWidget(self)
        self._stylus_modes_proxy.setWidget(self._stylus_modes_widget)
        self.scene().addItem(self._stylus_modes_proxy)
        self.set_stylus_proxy_pos()
        self.setAttribute(Qt.WA_TabletTracking)

        # # TOUCH GESTURES
        # recognizer = PanGestureRecognizer()
        # pan_gesture_id = QGestureRecognizer.registerRecognizer(recognizer) <--- CRASH HERE
        # self.grabGesture(pan_gesture_id)

        # DESIGN
        self.session.design.flow_theme_changed.connect(self._theme_changed)
        self.session.design.performance_mode_changed.connect(self._perf_mode_changed)

        # FRAMERATE TRACKING
        self.num_frames = 0
        self.framerate = 0
        self.framerate_timer = QTimer(self)
        self.framerate_timer.timeout.connect(self._on_framerate_timer_timeout)

        # self.show_framerate(m_sec_interval=100)  # for testing

        # CONFIG
        if config is not None:
            # # viewport update mode
            # vpum = config['viewport update mode']
            # if vpum == 'sync':  # vpum == VPUpdateMode.SYNC
            #     # self.vp_update_mode = VPUpdateMode.SYNC
            #     self.set_viewport_update_mode('sync')
            # elif vpum == 'async':  # vpum == VPUpdateMode.ASYNC
            #     # self.vp_update_mode = VPUpdateMode.ASYNC
            #     self.set_viewport_update_mode('async')

            if 'drawings' in config:  # not all (old) project files have drawings arr
                self.place_drawings_from_config(config['drawings'])

            if 'view size' in config:
                self.setSceneRect(0, 0, config['view size'][0], config['view size'][1])

            self._undo_stack.clear()

        # CATCH UP ON FLOW
        for node in self.flow.nodes:
            self.add_node(node)
        for c in self.flow.connections:
            self.add_connection(c)

    def show_framerate(self, show: bool = True, m_sec_interval: int = 1000):
        self._showing_framerate = show
        self.framerate_timer.setInterval(m_sec_interval)
        self.framerate_timer.start()

    def _on_framerate_timer_timeout(self):
        self.framerate = self.num_frames
        self.num_frames = 0

    def _init_shortcuts(self):
        place_new_node_shortcut = QShortcut(QKeySequence('Shift+P'), self)
        place_new_node_shortcut.activated.connect(self._place_new_node_by_shortcut)
        move_selected_components_left_shortcut = QShortcut(QKeySequence('Shift+Left'), self)
        move_selected_components_left_shortcut.activated.connect(self._move_selected_comps_left)
        move_selected_components_up_shortcut = QShortcut(QKeySequence('Shift+Up'), self)
        move_selected_components_up_shortcut.activated.connect(self._move_selected_comps_up)
        move_selected_components_right_shortcut = QShortcut(QKeySequence('Shift+Right'), self)
        move_selected_components_right_shortcut.activated.connect(self._move_selected_comps_right)
        move_selected_components_down_shortcut = QShortcut(QKeySequence('Shift+Down'), self)
        move_selected_components_down_shortcut.activated.connect(self._move_selected_comps_down)
        select_all_shortcut = QShortcut(QKeySequence('Ctrl+A'), self)
        select_all_shortcut.activated.connect(self.select_all)
        copy_shortcut = QShortcut(QKeySequence.Copy, self)
        copy_shortcut.activated.connect(self._copy)
        cut_shortcut = QShortcut(QKeySequence.Cut, self)
        cut_shortcut.activated.connect(self._cut)
        paste_shortcut = QShortcut(QKeySequence.Paste, self)
        paste_shortcut.activated.connect(self._paste)

        undo_shortcut = QShortcut(QKeySequence.Undo, self)
        undo_shortcut.activated.connect(self._undo_activated)
        redo_shortcut = QShortcut(QKeySequence.Redo, self)
        redo_shortcut.activated.connect(self._redo_activated)

    def _theme_changed(self, t):
        self._place_node_widget.setStyleSheet(self.session.design.node_selection_stylesheet)
        for n, ni in self.node_items.items():
            ni.widget.rebuild_ui()

        # TODO: repaint background. how?
        self.viewport().update()
        self.scene().update(self.sceneRect())

    def _perf_mode_changed(self, mode):

        # set widget value update rule for port inputs, because this takes up quite a bit of performance
        update_widget_value = (mode == 'pretty')
        for n, ni in self.node_items.items():
            for inp in ni.inputs:
                inp.update_widget_value = (update_widget_value and inp.widget)

        self.viewport().update()
        self.scene().update(self.sceneRect())

    def _scene_selection_changed(self):
        self.nodes_selection_changed.emit(self.selected_nodes())

    def contextMenuEvent(self, event):
        QGraphicsView.contextMenuEvent(self, event)
        # in the case of the menu already being shown by a widget under the mouse, the event is accepted here
        if event.isAccepted():
            return

        for i in self.items(event.pos()):
            if isinstance(i, NodeItem):
                ni: NodeItem = i
                menu: QMenu = ni.get_context_menu()
                menu.exec_(event.globalPos())
                event.accept()

    def _push_undo(self, cmd: FlowUndoCommand):
        self._undo_stack.push(cmd)
        cmd.activate()

    def _undo_activated(self):
        """Triggered by ctrl+z"""
        self._undo_stack.undo()
        self.viewport().update()

    def _redo_activated(self):
        """Triggered by ctrl+y"""
        self._undo_stack.redo()
        self.viewport().update()

    def mousePressEvent(self, event):
        # InfoMsgs.write('mouse press event received, point:', event.pos())

        # to catch tablet events (for some reason, it results in a mousePrEv too)
        if self.mouse_event_taken:
            self.mouse_event_taken = False
            return

        QGraphicsView.mousePressEvent(self, event)

        # don't process the event if it has been processed by a specific component in the scene
        # this includes node items, node port pins, proxy widgets etc.
        if self.mouse_event_taken:
            self.mouse_event_taken = False
            return

        if event.button() == Qt.LeftButton:
            if self._place_node_widget_proxy.isVisible():
                self.hide_place_node_widget()

        elif event.button() == Qt.RightButton:
            self._right_mouse_pressed_in_flow = True
            event.accept()

        self._mouse_press_pos = self.mapToScene(event.pos())

    def mouseMoveEvent(self, event):

        QGraphicsView.mouseMoveEvent(self, event)

        if self._right_mouse_pressed_in_flow:  # PAN

            if not self._panning:
                self._panning = True
                self._pan_last_x = event.x()
                self._pan_last_y = event.y()

            self.pan(event.pos())
            event.accept()

        self._last_mouse_move_pos = self.mapToScene(event.pos())

        if self._dragging_connection:
            self.viewport().repaint()

    def mouseReleaseEvent(self, event):
        # there might be a proxy widget meant to receive the event instead of the flow
        QGraphicsView.mouseReleaseEvent(self, event)

        node_item_at_event_pos = None
        for item in self.items(event.pos()):
            if isinstance(item, NodeItem):
                node_item_at_event_pos = item

        if self.mouse_event_taken:
            self.mouse_event_taken = False
            self.viewport().repaint()
            return

        elif self._panning:
            self._panning = False

        elif event.button() == Qt.RightButton:
            self._right_mouse_pressed_in_flow = False
            if self._mouse_press_pos == self._last_mouse_move_pos:
                self._place_node_widget.reset_list()
                self.show_place_node_widget(event.pos())
                return

        if self._dragging_connection:

            # connection dropped over port item
            port_items = {i: isinstance(i, PortItem) for i in self.items(event.pos())}
            if any(port_items.values()):

                p_i = list(port_items.keys())[list(port_items.values()).index(True)]  # gets the first PortItem

                # the validity of the connection is checked in connect_node_ports__cmd
                self.connect_node_ports__cmd(
                    self._selected_pin.port,
                    p_i.port
                )

            # connection dropped above NodeItem -> auto connect
            elif node_item_at_event_pos:
                # find node item
                ni_under_drop = None
                for item in self.items(event.pos()):
                    if isinstance(item, NodeItem):
                        ni_under_drop = item
                        self.auto_connect(self._selected_pin.port, ni_under_drop.node)
                        break

            # connection dropped somewhere else -> show node choice widget
            else:
                self._auto_connection_pin = self._selected_pin
                self.show_place_node_widget(event.pos())

            self._dragging_connection = False
            self._selected_pin = None

        # if event.button() == Qt.LeftButton:
        #     self._left_mouse_pressed_in_flow = False
        elif event.button() == Qt.RightButton:
            self._right_mouse_pressed_in_flow = False

        self.viewport().repaint()

    def keyPressEvent(self, event):
        QGraphicsView.keyPressEvent(self, event)

        if event.isAccepted():
            return

        if event.key() == Qt.Key_Escape:  # do I need that... ?
            self.clearFocus()
            self.setFocus()
            return True

        elif event.key() == Qt.Key_Delete:
            self.remove_selected_components__cmd()

    def wheelEvent(self, event):

        # ZOOM
        if event.modifiers() == Qt.CTRL and event.angleDelta().x() == 0:
            self.zoom(event.pos(), self.mapToScene(event.pos()), event.angleDelta().y())
            event.accept()
            return True

        QGraphicsView.wheelEvent(self, event)

    def tabletEvent(self, event):
        """tabletEvent gets called by stylus operations.
        LeftButton: std, no button pressed
        RightButton: upper button pressed"""

        # if in edit mode and not panning or starting a pan, pass on to std mouseEvent handlers above
        if self.stylus_mode == 'edit' and not self._panning and not \
                (event.type() == QTabletEvent.TabletPress and event.button() == Qt.RightButton):
            return  # let the mousePress/Move/Release-Events handle it

        scaled_event_pos: QPointF = event.posF() / self._current_scale

        if event.type() == QTabletEvent.TabletPress:
            self.mouse_event_taken = True

            if event.button() == Qt.LeftButton:
                if self.stylus_mode == 'comment':
                    view_pos = self.mapToScene(self.viewport().pos())
                    new_drawing = self._create_and_place_drawing__cmd(
                        view_pos + scaled_event_pos,
                        config={**self._stylus_modes_widget.get_pen_settings(), 'viewport pos': view_pos}
                    )
                    self._current_drawing = new_drawing
                    self._drawing = True
            elif event.button() == Qt.RightButton:
                self._panning = True
                self._pan_last_x = event.x()
                self._pan_last_y = event.y()

        elif event.type() == QTabletEvent.TabletMove:
            self.mouse_event_taken = True
            if self._panning:
                self.pan(event.pos())

            elif event.pointerType() == QTabletEvent.Eraser:
                if self.stylus_mode == 'comment':
                    for i in self.items(event.pos()):
                        if isinstance(i, DrawingObject):
                            self.remove_drawing(i)
                            break
            elif self.stylus_mode == 'comment' and self._drawing:
                if self._current_drawing.append_point(scaled_event_pos):
                    self._current_drawing.stroke_weights.append(event.pressure())
                self._current_drawing.update()
                self.viewport().update()

        elif event.type() == QTabletEvent.TabletRelease:
            if self._panning:
                self._panning = False
            if self.stylus_mode == 'comment' and self._drawing:
                self._current_drawing.finish()
                InfoMsgs.write('drawing finished')
                self._current_drawing = None
                self._drawing = False

    """
    --> https://forum.qt.io/topic/121473/qgesturerecognizer-registerrecognizer-crashes-using-pyside2

    def event(self, event) -> bool:
        # if event.type() == QEvent.Gesture:
        #     if event.gesture(PanGesture) is not None:
        #         return self.pan_gesture(event)

        return QGraphicsView.event(self, event)

    def pan_gesture(self, event: QGestureEvent) -> bool:
        pan: PanGesture = event.gesture(PanGesture)
        print(pan)
        return True
    """

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat('text/plain'):
            event.acceptProposedAction()

    def dragMoveEvent(self, event):
        if event.mimeData().hasFormat('text/plain'):
            event.acceptProposedAction()

    # def dropEvent(self, event):
    #     text = event.mimeData().text()
    #     item: QListWidgetItem = event.mimeData()
    #     InfoMsgs.write('drop received in Flow:', text)
    #
    #     j_obj = None
    #     type = ''
    #     try:
    #         j_obj = json.loads(text)
    #         type = j_obj['type']
    #     except Exception:
    #         return
    #
    #     if type == 'variable':
    #         self.show_node_choice_widget(event.pos(),  # only show get_var and set_var nodes
    #                                      [n for n in self.session.nodes if find_type_in_object(n, GetVar_Node) or
    #                                       find_type_in_object(n, SetVar_Node)])

    # PAINTING
    def drawBackground(self, painter, rect):

        painter.setBrush(self.session.design.flow_theme.flow_background_brush)
        painter.drawRect(rect.intersected(self.sceneRect()))
        painter.setPen(Qt.NoPen)
        painter.drawRect(self.sceneRect())

        if self.session.design.performance_mode == 'pretty':
            theme = self.session.design.flow_theme
            if theme.flow_background_grid and self._current_scale >= 0.7:
                if theme.flow_background_grid[0] == 'points':
                    color = theme.flow_background_grid[1]
                    pen_width = theme.flow_background_grid[2]
                    diff_x = theme.flow_background_grid[3]
                    diff_y = theme.flow_background_grid[4]

                    pen = QPen(color)
                    pen.setWidthF(pen_width)
                    painter.setPen(pen)

                    for x in range(diff_x, self.sceneRect().toRect().width(), diff_x):
                        for y in range(diff_y, self.sceneRect().toRect().height(), diff_y):
                            painter.drawPoint(x, y)

        self.set_stylus_proxy_pos()  # has to be called here instead of in drawForeground to prevent lagging
        # self.set_zoom_proxy_pos()

    def drawForeground(self, painter, rect):

        if self._showing_framerate:
            self.num_frames += 1
            pen = QPen(QColor('#A9D5EF'))
            pen.setWidthF(2)
            painter.setPen(pen)

            pos = self.mapToScene(10, 23)
            painter.setFont(QFont('Poppins', round(11 * self._total_scale_div)))
            painter.drawText(pos, "{:.2f}".format(self.framerate))

        # DRAW CURRENTLY DRAGGED CONNECTION
        if self._dragging_connection:
            pen = QPen('#101520')
            pen.setWidth(3)
            pen.setStyle(Qt.DotLine)
            painter.setPen(pen)

            pin_pos = self._selected_pin.get_scene_center_pos()
            spp = self._selected_pin.port
            cursor_pos = self._last_mouse_move_pos

            pos1 = pin_pos if spp.io_pos == PortObjPos.OUTPUT else cursor_pos
            pos2 = pin_pos if spp.io_pos == PortObjPos.INPUT else cursor_pos

            painter.drawPath(
                default_cubic_connection_path(pos1, pos2)
            )

        # DRAW SELECTED NIs BORDER
        for ni in self.selected_node_items():
            pen = QPen(self.session.design.flow_theme.flow_highlight_pen_color)
            pen.setWidth(3)
            painter.setPen(pen)
            painter.setBrush(Qt.NoBrush)

            size_factor = 1.2
            x = ni.pos().x() - ni.boundingRect().width() / 2 * size_factor
            y = ni.pos().y() - ni.boundingRect().height() / 2 * size_factor
            w = ni.boundingRect().width() * size_factor
            h = ni.boundingRect().height() * size_factor
            painter.drawRoundedRect(x, y, w, h, 10, 10)

        # DRAW SELECTED DRAWINGS BORDER
        for p_o in self.selected_drawings():
            pen = QPen(QColor('#a3cc3b'))
            pen.setWidth(2)
            painter.setPen(pen)
            painter.setBrush(Qt.NoBrush)

            size_factor = 1.05
            x = p_o.pos().x() - p_o.width / 2 * size_factor
            y = p_o.pos().y() - p_o.height / 2 * size_factor
            w = p_o.width * size_factor
            h = p_o.height * size_factor
            painter.drawRoundedRect(x, y, w, h, 6, 6)
            painter.drawEllipse(p_o.pos().x(), p_o.pos().y(), 2, 2)

    def get_viewport_img(self) -> QImage:
        """Returns a clear image of the viewport"""

        self.hide_proxies()
        img = QImage(self.viewport().rect().width(), self.viewport().height(), QImage.Format_ARGB32)
        img.fill(Qt.transparent)

        painter = QPainter(img)
        painter.setRenderHint(QPainter.Antialiasing)
        self.render(painter, self.viewport().rect(), self.viewport().rect())
        self.show_proxies()
        return img

    def get_whole_scene_img(self) -> QImage:
        """Returns an image of the whole scene, scaled accordingly to current scale factor.
        Due to a bug this only works from the viewport position down and right, so the user has to scroll to
        the top left corner in order to get the full scene"""

        self.hide_proxies()
        img = QImage(self.sceneRect().width() / self._total_scale_div,
                     self.sceneRect().height() / self._total_scale_div,
                     QImage.Format_RGB32)
        img.fill(Qt.transparent)

        painter = QPainter(img)
        painter.setRenderHint(QPainter.Antialiasing)
        rect = QRectF()
        rect.setLeft(-self.viewport().pos().x())
        rect.setTop(-self.viewport().pos().y())
        rect.setWidth(img.rect().width())
        rect.setHeight(img.rect().height())
        # rect is right... but it only renders from the viewport's point down-and rightwards, not from topleft (0,0) ...
        self.render(painter, rect, rect.toRect())
        self.show_proxies()
        return img

    # PROXY POSITIONS

    # def set_zoom_proxy_pos(self):
    #     self._zoom_proxy.setPos(self.mapToScene(self.viewport().width() - self._zoom_widget.width(), 0))

    def set_stylus_proxy_pos(self):
        self._stylus_modes_proxy.setPos(
            self.mapToScene(self.viewport().width() - self._stylus_modes_widget.width(), 0))
            # self.mapToScene(self.viewport().width() - self._stylus_modes_widget.width() - self._zoom_widget.width(), 0))

    def hide_proxies(self):
        self._stylus_modes_proxy.hide()
        # self._zoom_proxy.hide()

    def show_proxies(self):
        self._stylus_modes_proxy.show()
        # self._zoom_proxy.show()

    # PLACE NODE WIDGET
    def show_place_node_widget(self, pos, nodes=None):
        """Opens the place node dialog in the scene."""

        # calculating position
        self._node_place_pos = self.mapToScene(pos)
        dialog_pos = QPoint(pos.x() + 1, pos.y() + 1)

        # ensure that the node_selection_widget stays in the viewport
        if dialog_pos.x() + self._place_node_widget.width() / self._total_scale_div > self.viewport().width():
            dialog_pos.setX(dialog_pos.x() - (
                    dialog_pos.x() + self._place_node_widget.width() / self._total_scale_div - self.viewport().width()))
        if dialog_pos.y() + self._place_node_widget.height() / self._total_scale_div > self.viewport().height():
            dialog_pos.setY(dialog_pos.y() - (
                    dialog_pos.y() + self._place_node_widget.height() / self._total_scale_div - self.viewport().height()))
        dialog_pos = self.mapToScene(dialog_pos)

        # open nodes dialog
        # the dialog emits 'node_chosen' which is connected to self.place_node,
        # so this all continues at self.place_node below
        self._place_node_widget.update_list(nodes if nodes is not None else self.session.nodes)
        self._place_node_widget.update_view()
        self._place_node_widget_proxy.setPos(dialog_pos)
        self._place_node_widget_proxy.show()
        self._place_node_widget.refocus()

    def hide_place_node_widget(self):
        self._place_node_widget_proxy.hide()
        self._place_node_widget.clearFocus()
        self._auto_connection_pin = None

    def _place_new_node_by_shortcut(self):  # Shift+P
        point_in_viewport = None
        selected_NIs = self.selected_node_items()
        if len(selected_NIs) > 0:
            x = selected_NIs[-1].pos().x() + 150
            y = selected_NIs[-1].pos().y()
            self._node_place_pos = QPointF(x, y)
            point_in_viewport = self.mapFromScene(QPoint(x, y))
        else:  # place in center
            viewport_x = self.viewport().width() / 2
            viewport_y = self.viewport().height() / 2
            point_in_viewport = QPointF(viewport_x, viewport_y).toPoint()
            self._node_place_pos = self.mapToScene(point_in_viewport)

        self._place_node_widget.reset_list()
        self.show_place_node_widget(point_in_viewport)

    # PAN
    def pan(self, new_pos):
        self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - (new_pos.x() - self._pan_last_x))
        self.verticalScrollBar().setValue(self.verticalScrollBar().value() - (new_pos.y() - self._pan_last_y))
        self._pan_last_x = new_pos.x()
        self._pan_last_y = new_pos.y()

    # ZOOM
    def zoom_in(self, amount):
        local_viewport_center = QPoint(self.viewport().width() / 2, self.viewport().height() / 2)
        self.zoom(local_viewport_center, self.mapToScene(local_viewport_center), amount)

    def zoom_out(self, amount):
        local_viewport_center = QPoint(self.viewport().width() / 2, self.viewport().height() / 2)
        self.zoom(local_viewport_center, self.mapToScene(local_viewport_center), -amount)

    def zoom(self, p_abs, p_mapped, angle):
        by = 0
        velocity = 2 * (1 / self._current_scale) + 0.5
        if velocity > 3:
            velocity = 3

        direction = ''
        if angle > 0:
            by = 1 + (angle / 360 * 0.1 * velocity)
            direction = 'in'
        elif angle < 0:
            by = 1 - (-angle / 360 * 0.1 * velocity)
            direction = 'out'
        else:
            by = 1

        scene_rect_width = self.mapFromScene(self.sceneRect()).boundingRect().width()
        scene_rect_height = self.mapFromScene(self.sceneRect()).boundingRect().height()

        if direction == 'in':
            if self._current_scale * by < 3:
                self.scale(by, by)
                self._current_scale *= by
        elif direction == 'out':
            if scene_rect_width * by >= self.viewport().size().width() and scene_rect_height * by >= self.viewport().size().height():
                self.scale(by, by)
                self._current_scale *= by

        w = self.viewport().width()
        h = self.viewport().height()
        wf = self.mapToScene(QPoint(w - 1, 0)).x() - self.mapToScene(QPoint(0, 0)).x()
        hf = self.mapToScene(QPoint(0, h - 1)).y() - self.mapToScene(QPoint(0, 0)).y()
        lf = p_mapped.x() - p_abs.x() * wf / w
        tf = p_mapped.y() - p_abs.y() * hf / h

        self.ensureVisible(lf, tf, wf, hf, 0, 0)

        target_rect = QRectF(QPointF(lf, tf),
                             QSizeF(wf, hf))
        self._total_scale_div = target_rect.width() / self.viewport().width()

        self.ensureVisible(target_rect, 0, 0)

    # NODES
    def create_node__cmd(self, node_class):
        self._push_undo(
            PlaceNode_Command(self, node_class, self._node_place_pos)
        )

    def add_node(self, node):

        # create item
        item: NodeItem = None

        if node in self.node_items__cache.keys():  # load from cache
            # print('using a cached item')
            item = self.node_items__cache[node]
            self._add_node_item(item)

        else:  # create new item
            item_config = node.init_config
            item = NodeItem(node, params=(self, self.session.design, item_config))
            node.item = item
            item.initialize()
            self.node_placed.emit(node)

            pos = None
            if item_config is not None:
                if 'pos x' in item_config:
                    pos = QPointF(item_config['pos x'], item_config['pos y'])
                elif 'position x' in item_config:  # backwards compatibility
                    pos = QPointF(item_config['position x'], item_config['position y'])
            else:
                pos = self._node_place_pos

            self._add_node_item(item, pos)

        # auto connect
        if self._auto_connection_pin:
            self.auto_connect(self._auto_connection_pin.port,
                              node)

    def _add_node_item(self, item: NodeItem, pos=None):
        self.node_items[item.node] = item

        self.scene().addItem(item)
        if pos:
            item.setPos(pos)

        # select new item
        self.clear_selection()
        item.setSelected(True)

    def remove_node(self, node):
        item = self.node_items[node]
        self._remove_node_item(item)
        del self.node_items[node]

    def _remove_node_item(self, item: NodeItem):
        # store item in case the remove action gets undone later
        self.node_items__cache[item.node] = item
        self.scene().removeItem(item)

    # CONNECTIONS
    def connect_node_ports__cmd(self, p1: NodePort, p2: NodePort):
        self._temp_connection_ports = (p1, p2)
        self.check_connection_validity_request.emit(p1, p2, True)

    def connection_request_valid(self, valid: bool):
        """
        Triggered after the abstract flow evaluated validity of pending connect request.
        This can also lead to a disconnect!
        """

        if valid:
            out, inp = self._temp_connection_ports
            if out.io_pos == PortObjPos.INPUT:
                out, inp = inp, out

            # remove forbidden connections
            if inp.type_ == 'data':
                for c in inp.connections:

                    if c.out == out:
                        # if the exact connection exists, we want to remove it by command
                        continue

                    self._push_undo(
                        ConnectPorts_Command(self, out=c.out, inp=inp)
                    )

            self._push_undo(
                ConnectPorts_Command(self, out=out, inp=inp)
            )

    def add_connection(self, c: Connection):

        item: ConnectionItem = None
        if c in self.connection_items__cache.keys():
            item = self.connection_items__cache[c]

        else:
            if isinstance(c, DataConnection):
                item = CLASSES['data conn item'](c, self.session.design)
            else:
                item = CLASSES['exec conn item'](c, self.session.design)

        self._add_connection_item(item)

        item.out_item.port_connected()
        item.inp_item.port_connected()

    def _add_connection_item(self, item: ConnectionItem):
        self.connection_items[item.connection] = item
        self.scene().addItem(item)
        item.setZValue(10)
        # self.viewport().repaint()

    def remove_connection(self, c: Connection):
        item = self.connection_items[c]
        self._remove_connection_item(item)

        item.out_item.port_disconnected()
        item.inp_item.port_disconnected()

        del self.connection_items[c]

    def _remove_connection_item(self, item: ConnectionItem):
        self.connection_items__cache[item.connection] = item
        self.scene().removeItem(item)

    def auto_connect(self, p: NodePort, n: Node):
        if p.io_pos == PortObjPos.OUTPUT:
            for inp in n.inputs:
                if p.type_ == inp.type_:
                    # connect exactly once
                    self.connect_node_ports__cmd(p, inp)
                    return
        elif p.io_pos == PortObjPos.INPUT:
            for out in n.outputs:
                if p.type_ == out.type_:
                    # connect exactly once
                    self.connect_node_ports__cmd(p, out)
                    return

    def update_conn_item(self, c: Connection):
        if c in self.connection_items:
            self.connection_items[c].changed = True
            self.connection_items[c].update()

    # DRAWINGS
    def create_drawing(self, config=None) -> DrawingObject:
        """Creates and returns a new DrawingObject."""

        new_drawing = DrawingObject(self, config)
        return new_drawing

    def add_drawing(self, drawing_obj, posF=None):
        """Adds a DrawingObject to the scene."""

        self.scene().addItem(drawing_obj)
        if posF:
            drawing_obj.setPos(posF)
        self.drawings.append(drawing_obj)

    def add_drawings(self, drawings):
        """Adds a list of DrawingObjects to the scene."""

        for d in drawings:
            self.add_drawing(d)

    def remove_drawing(self, drawing: DrawingObject):
        """Removes a drawing from the scene."""

        # TODO https://github.com/leon-thomm/ryvencore/issues/4

        self.scene().removeItem(drawing)
        self.drawings.remove(drawing)

    def place_drawings_from_config(self, drawings_config: list, offset_pos=QPoint(0, 0)):
        """Creates and places drawings from drawings. The same list is returned by the config_data() method
        at 'drawings'."""

        new_drawings = []
        for d_config in drawings_config:
            x = d_config['pos x'] + offset_pos.x()
            y = d_config['pos y'] + offset_pos.y()
            new_drawing = self.create_drawing(config=d_config)
            self.add_drawing(new_drawing, QPointF(x, y))
            new_drawings.append(new_drawing)

        return new_drawings

    def _create_and_place_drawing__cmd(self, posF, config=None):
        new_drawing_obj = self.create_drawing(config)
        place_command = PlaceDrawing_Command(self, posF, new_drawing_obj)
        self._push_undo(place_command)
        return new_drawing_obj

    # ADDING/REMOVING COMPONENTS
    def add_component(self, e: QGraphicsItem):
        if isinstance(e, NodeItem):
            self.add_node(e.node)
            self.add_node_item(e)
        elif isinstance(e, DrawingObject):
            self.add_drawing(e)

    def remove_components(self, comps: [QGraphicsItem]):
        for c in comps:
            self.remove_component(c)

    def remove_component(self, e: QGraphicsItem):
        if isinstance(e, NodeItem):
            self.remove_node(e.node)
            self.remove_node_item(e)
        elif isinstance(e, DrawingObject):
            self.remove_drawing(e)

    def remove_selected_components__cmd(self):
        self._push_undo(
            RemoveComponents_Command(self, self.scene().selectedItems())
        )

        self.viewport().update()

    # MOVING COMPONENTS
    def _move_selected_copmonents__cmd(self, x, y):
        new_rel_pos = QPointF(x, y)

        # if one node item would leave the scene (f.ex. pos.x < 0), stop
        left = False
        for i in self.scene().selectedItems():
            new_pos = i.pos() + new_rel_pos
            w = i.boundingRect().width()
            h = i.boundingRect().height()
            if new_pos.x() - w / 2 < 0 or \
                    new_pos.x() + w / 2 > self.scene().width() or \
                    new_pos.y() - h / 2 < 0 or \
                    new_pos.y() + h / 2 > self.scene().height():
                left = True
                break

        if not left:
            # moving the items
            items_group = self.scene().createItemGroup(self.scene().selectedItems())
            items_group.moveBy(new_rel_pos.x(), new_rel_pos.y())
            self.scene().destroyItemGroup(items_group)

            # saving the command
            self._push_undo(
                MoveComponents_Command(self, self.scene().selectedItems(), p_from=-new_rel_pos, p_to=QPointF(0, 0))
            )

        self.viewport().repaint()

    def _move_selected_comps_left(self):
        self._move_selected_copmonents__cmd(-40, 0)

    def _move_selected_comps_up(self):
        self._move_selected_copmonents__cmd(0, -40)

    def _move_selected_comps_right(self):
        self._move_selected_copmonents__cmd(+40, 0)

    def _move_selected_comps_down(self):
        self._move_selected_copmonents__cmd(0, +40)

    # SELECTION
    def selected_components_moved(self, pos_diff):
        items_list = self.scene().selectedItems()

        self._push_undo(
            MoveComponents_Command(self, items_list, p_from=-pos_diff, p_to=QPointF(0, 0))
        )

    def selected_node_items(self) -> [NodeItem]:
        """Returns a list of the currently selected NodeItems."""

        selected_NIs = []
        for i in self.scene().selectedItems():
            if isinstance(i, NodeItem):
                selected_NIs.append(i)
        return selected_NIs

    def selected_nodes(self) -> [Node]:
        return [item.node for item in self.selected_node_items()]

    def selected_drawings(self) -> [DrawingObject]:
        """Returns a list of the currently selected drawings."""

        selected_drawings = []
        for i in self.scene().selectedItems():
            if isinstance(i, DrawingObject):
                selected_drawings.append(i)
        return selected_drawings

    def select_all(self):
        for i in self.scene().items():
            if i.ItemIsSelectable:
                i.setSelected(True)
        self.viewport().repaint()

    def clear_selection(self):
        self.scene().clearSelection()

    def select_components(self, comps):
        self.scene().clearSelection()
        for c in comps:
            c.setSelected(True)

    # ACTIONS
    def _copy(self):  # ctrl+c
        data = {'nodes': self._get_nodes_config_data(self.selected_nodes()),
                'connections': self._get_connections_config_data(self.selected_nodes()),
                'drawings': self._get_drawings_config_data(self.selected_drawings())}
        QGuiApplication.clipboard().setText(json.dumps(data))

    def _cut(self):  # ctrl+x
        data = {'nodes': self._get_nodes_config_data(self.selected_nodes()),
                'connections': self._get_connections_config_data(self.selected_nodes()),
                'drawings': self._get_drawings_config_data(self.selected_drawings())}
        QGuiApplication.clipboard().setText(json.dumps(data))
        self.remove_selected_components__cmd()

    def _paste(self):
        data = {}
        try:
            data = json.loads(QGuiApplication.clipboard().text())
        except Exception as e:
            return

        self.clear_selection()

        # calculate offset
        positions = []
        for d in data['drawings']:
            positions.append({'x': d['pos x'],
                              'y': d['pos y']})
        for n in data['nodes']:
            positions.append({'x': n['pos x'],
                              'y': n['pos y']})

        offset_for_middle_pos = QPointF(0, 0)
        if len(positions) > 0:
            rect = QRectF(positions[0]['x'], positions[0]['y'], 0, 0)
            for p in positions:
                x = p['x']
                y = p['y']
                if x < rect.left():
                    rect.setLeft(x)
                if x > rect.right():
                    rect.setRight(x)
                if y < rect.top():
                    rect.setTop(y)
                if y > rect.bottom():
                    rect.setBottom(y)

            offset_for_middle_pos = self._last_mouse_move_pos - rect.center()

        self._push_undo(
            Paste_Command(self, data, offset_for_middle_pos)
        )

    # # VIEWPORT UPDATE MODE
    # def viewport_update_mode(self) -> str:
    #     """Returns the current viewport update mode as string (sync or async) of the flow"""
    #     return VPUpdateMode.stringify(self.vp_update_mode)
    #
    # def set_viewport_update_mode(self, mode: str):
    #     """
    #     Sets the viewport update mode of the flow
    #     :mode: 'sync' or 'async'
    #     """
    #     if mode == 'sync':
    #         self.vp_update_mode = VPUpdateMode.SYNC
    #     elif mode == 'async':
    #         self.vp_update_mode = VPUpdateMode.ASYNC
    #
    #     self.viewport_update_mode_changed.emit(self.viewport_update_mode())

    # CONFIG
    def generate_config_data(self, script_data):

        flow_config, nodes_data, connections_data = script_data['flow']

        script_data['flow'] = {
            **flow_config,
            'nodes': self.complete_nodes_config_data(nodes_data),
            'connections': self.complete_connections_config_data(connections_data),
        }

        script_data['flow view'] = {
            'drawings': self._get_drawings_config_data(self.drawings),
            'view size': [self.sceneRect().size().width(), self.sceneRect().size().height()]
        }

        self._tmp_data = script_data

        return script_data

    def complete_nodes_config_data(self, nodes_config_dict):
        """
        Adds the item config (scene pos etc.) to the config of the nodes.
        """

        nodes_config_list = []
        for n in list(nodes_config_dict.keys()):
            cfg = nodes_config_dict[n]

            item = self.node_items[n]
            nodes_config_list.append(item.complete_config(cfg))

        return nodes_config_list

    def complete_connections_config_data(self, conn_config_dict):
        """
        Adds the item config to the config of the connections
        """

        # TODO: add custom config for connection items

        conns_config_list = []
        for c in list(conn_config_dict.keys()):
            conns_config_list.append(conn_config_dict[c])

        return conns_config_list

    def _get_nodes_config_data(self, nodes):

        # wait for abstract flow
        self.flow._tmp_data = None
        self.get_nodes_config_request.emit(nodes)
        while self.flow._tmp_data is None:
            time.sleep(0.001)

        return self.complete_nodes_config_data(self.flow._tmp_data)

    def _get_connections_config_data(self, nodes):

        # wait for abstract flow
        self.flow._tmp_data = None
        self.get_connections_config_request.emit(nodes)
        while self.flow._tmp_data is None:
            time.sleep(0.001)

        return self.complete_connections_config_data(self.flow._tmp_data)

    def _get_drawings_config_data(self, drawings):
        drawings_list = []
        for drawing in drawings:
            drawing_dict = drawing.config_data()

            drawings_list.append(drawing_dict)

        return drawings_list



from logging import Logger as PyLogger


class Logger(PyLogger):
    """
    A small wrapper template for the python Loggers to add
    functionality on node events.
    Reimplemented as wrapper by the frontend with
    according implementations of the below methods.
    """

    def enable(self):
        pass

    def disable(self):
        pass



from qtpy.QtWidgets import QGraphicsItem, QMenu, QGraphicsDropShadowEffect
from qtpy.QtCore import Qt, QRectF, QObject, QPointF
from qtpy.QtGui import QColor

from .ryvencore.NodePort import NodeInput, NodeOutput
from .NodeItemAction import NodeItemAction
from .NodeItemAnimator import NodeItemAnimator
from .NodeItemWidget import NodeItemWidget
from .PortItem import InputPortItem, OutputPortItem
from .ryvencore.tools import serialize, deserialize
from .tools import MovementEnum


class NodeItem(QGraphicsItem, QObject):
    """The GUI representative for nodes. Unlike the Node class, this class is not subclassed individually and works
    the same for every node."""

    def __init__(self, node, params):
        QGraphicsItem.__init__(self)
        QObject.__init__(self)

        self.node = node
        flow_view, design, config = params
        self.flow_view = flow_view
        self.session_design = design
        self.movement_state = None
        self.movement_pos_from = None
        self.painted_once = False
        self.inputs = []
        self.outputs = []
        self.color = QColor(self.node.color)  # manipulated by self.animator

        self.collapsed = False
        self.hovered = False
        self.hiding_unconnected_ports = False

        self.personal_logs = []

        # 'initializing' will be set to False below. It's needed for the ports setup, to prevent shape updating stuff
        self.initializing = True

        # self.temp_state_data = None
        self.init_config = config

        # CONNECT TO NODE
        self.node.updated.connect(self.node_updated)
        self.node.update_shape_triggered.connect(self.update_shape)
        self.node.hide_unconnected_ports_triggered.connect(self.hide_unconnected_ports_triggered)
        self.node.show_unconnected_ports_triggered.connect(self.show_unconnected_ports_triggered)
        self.node.input_added.connect(self.add_new_input)
        self.node.output_added.connect(self.add_new_output)
        self.node.input_removed.connect(self.remove_input)
        self.node.output_removed.connect(self.remove_output)

        # FLAGS
        self.setFlags(QGraphicsItem.ItemIsSelectable | QGraphicsItem.ItemIsMovable |
                      QGraphicsItem.ItemSendsScenePositionChanges)
        self.setAcceptHoverEvents(True)
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)

        # UI
        self.shadow_effect = None
        self.main_widget = None
        if self.node.main_widget_class is not None:
            self.main_widget = self.node.main_widget_class((self.node, self))
        self.widget = NodeItemWidget(self.node, self)  # QGraphicsWidget(self)
        self.animator = NodeItemAnimator(self)  # needs self.title_label

        # TOOLTIP
        if self.node.description_html:
            self.setToolTip(self.node.description_html)
        elif self.node.__doc__:
            self.setToolTip('<html><head/><body><p>' + self.node.__doc__ + '</p></body></html>')
        self.setCursor(Qt.SizeAllCursor)

        # DESIGN THEME
        self.session_design.flow_theme_changed.connect(self.update_design)
        self.session_design.performance_mode_changed.connect(self.update_design)

    def initialize(self):
        """All ports and the main widget get finally created here."""

        # LOADING CONFIG
        if self.init_config is not None:
            if self.main_widget:
                try:
                    self.main_widget.set_state(deserialize(self.init_config['main widget data']))
                except Exception as e:
                    print('Exception while setting data in', self.node.title, 'Node\'s main widget:', e,
                          ' (was this intended?)')

        # catch up on ports
        for i in self.node.inputs:
            self.add_new_input(i)

        for o in self.node.outputs:
            self.add_new_output(o)

        if self.init_config is not None:
            if self.init_config.get('unconnected ports hidden'):
                self.hide_unconnected_ports_triggered()
            if self.init_config.get('collapsed'):
                self.collapse()

        self.initializing = False

        # No self.update_shape() here because for some reason, the bounding rect hasn't been initialized yet, so
        # self.update_shape() gets called when the item is being drawn the first time (see paint event in NI painter)
        # https://forum.qt.io/topic/117179/force-qgraphicsitem-to-update-immediately-wait-for-update-event

        self.update_design()  # load current design, update QGraphicsItem

        self.update()  # ... not sure if I need that

    # --------------------------------------------------------------------------------------
    # UI STUFF ----------------------------------------

    def node_updated(self):
        if self.session_design.animations_enabled:
            if not self.animator.running():
                self.animator.start()
            elif self.animator.fading_out():
                self.animator.set_animation_max()

        self.update()

    def add_new_input(self, inp: NodeInput, insert: int = None):

        # create item
        # inp.item = InputPortItem(inp.node, self, inp)
        item = InputPortItem(inp.node, self, inp)

        if insert is not None:
            self.inputs.insert(insert, item)
            self.widget.insert_input_into_layout(insert, item)
        else:
            self.inputs.append(item)
            self.widget.add_input_to_layout(item)

        if not self.initializing:
            self.update_shape()
            self.update()

    def remove_input(self, inp: NodeInput):
        item = None
        for inp_item in self.inputs:
            if inp_item.port == inp:
                item = inp_item
                break

        # index = self.node.inputs.index(inp)
        # item = self.inputs[index]

        # for some reason, I have to remove all widget items manually from the scene too. setting the items to
        # ownedByLayout(True) does not work, I don't know why.
        self.scene().removeItem(item.pin)
        self.scene().removeItem(item.label)
        if item.proxy is not None:
            self.scene().removeItem(item.proxy)

        self.inputs.remove(item)
        self.widget.remove_input_from_layout(item)

        if not self.initializing:
            self.update_shape()
            self.update()

    def add_new_output(self, out: NodeOutput, insert: int = None):

        # create item
        # out.item = OutputPortItem(out.node, self, out)
        item = OutputPortItem(out.node, self, out)

        if insert is not None:
            self.outputs.insert(insert, item)
            self.widget.insert_output_into_layout(insert, item)
        else:
            self.outputs.append(item)
            self.widget.add_output_to_layout(item)

        if not self.initializing:
            self.update_shape()
            self.update()

    def remove_output(self, out: NodeOutput):
        item = None
        for out_item in self.outputs:
            if out_item.port == out:
                item = out_item
                break

        # index = self.node.outputs.index(out)
        # item = self.outputs[index]

        # see remove_input() for info!
        self.scene().removeItem(item.pin)
        self.scene().removeItem(item.label)

        self.outputs.remove(item)
        self.widget.remove_output_from_layout(item)

        if not self.initializing:
            self.update_shape()
            self.update()

    def update_shape(self):
        self.widget.update_shape()
        self.update_conn_pos()
        self.flow_view.viewport().update()

    def update_design(self):
        """Loads the shadow effect option and causes redraw with active theme."""

        if self.session_design.node_item_shadows_enabled:
            self.shadow_effect = QGraphicsDropShadowEffect()
            self.shadow_effect.setXOffset(12)
            self.shadow_effect.setYOffset(12)
            self.shadow_effect.setBlurRadius(20)
            self.shadow_effect.setColor(self.session_design.flow_theme.node_item_shadow_color)
            self.setGraphicsEffect(self.shadow_effect)
        else:
            self.setGraphicsEffect(None)

        self.widget.update_shape()
        self.animator.reload_values()

        QGraphicsItem.update(self)

    def boundingRect(self):
        # remember: (0, 0) shall be the NI's center!
        rect = QRectF()
        w = self.widget.layout().geometry().width()
        h = self.widget.layout().geometry().height()
        rect.setLeft(-w / 2)
        rect.setTop(-h / 2)
        rect.setWidth(w)
        rect.setHeight(h)
        return rect

    def get_left_body_header_vertex_scene_pos(self):
        return self.mapToScene(
            QPointF(
                -self.boundingRect().width() / 2,
                -self.boundingRect().height() / 2 + self.widget.header_widget.rect().height()
            )
        )

    def get_right_body_header_vertex_scene_pos(self):
        return self.mapToScene(
            QPointF(
                +self.boundingRect().width() / 2,
                -self.boundingRect().height() / 2 + self.widget.header_widget.rect().height()
            )
        )

    def hide_unconnected_ports_triggered(self):
        self.widget.hide_unconnected_ports()
        self.hiding_unconnected_ports = True
        self.update_shape()

    def show_unconnected_ports_triggered(self):
        self.widget.show_unconnected_ports()
        self.hiding_unconnected_ports = False
        self.update_shape()

    def expand(self):
        self.collapsed = False
        self.widget.expand()
        self.update_shape()

    def collapse(self):
        self.collapsed = True
        self.widget.collapse()
        self.update_shape()

    #   PAINTING

    def paint(self, painter, option, widget=None):
        """All painting is done by NodeItemPainter"""

        # in order to access a meaningful geometry of GraphicsWidget contents in update_shape(), the paint event
        # has to be called once. See here:
        # https://forum.qt.io/topic/117179/force-qgraphicsitem-to-update-immediately-wait-for-update-event/4
        if not self.painted_once:
            # ok, quick notice. Since I am using a NodeItemWidget, calling self.update_design() here (again)
            # leads to a QT crash without error, which is really strange. Calling update_design multiple times
            # principally isn't a problem, but, for some reason, here it leads to a crash in QT. It's not necessary
            # anymore, so I removed it.
            # self.update_design()

            self.update_shape()
            self.update_conn_pos()

        self.session_design.flow_theme.paint_NI(
            node=self.node,
            node_style=self.node.style,
            painter=painter,
            option=option,
            color=self.color,
            w=self.boundingRect().width(),
            h=self.boundingRect().height(),
            bounding_rect=self.boundingRect(),
            title_rect=self.widget.header_widget.boundingRect()
            if self.widget.header_widget
            else self.widget.title_label.boundingRect()
        )

        # useful for widget development:

        # painter.setBrush(Qt.NoBrush)
        # painter.setPen(QPen(QColor('black')))
        #
        # painter.drawRect(
        #         self.boundingRect()
        # )
        #
        # header_rect = QRectF(
        #         -self.boundingRect().width()/2, -self.boundingRect().height()/2,
        #         self.widget.header_widget.geometry().width(), self.widget.header_widget.geometry().height()
        #     )
        # painter.drawRect(header_rect)
        #
        # # note that the widget's layout has spacing 0
        # body_rect = QRectF(
        #     -self.boundingRect().width()/2, -self.boundingRect().height()/2 + header_rect.height(),
        #     self.widget.body_widget.geometry().width(), self.widget.body_widget.geometry().height()
        # )
        # painter.drawRect(body_rect)

        self.painted_once = True

    # MOUSE INTERACTION

    def get_context_menu(self):
        menu = QMenu(self.flow_view)

        for a in self.get_actions(self.node.get_extended_default_actions(), menu):  # menu needed for 'parent'
            if type(a) == NodeItemAction:
                menu.addAction(a)
            elif type(a) == QMenu:
                menu.addMenu(a)

        menu.addSeparator()

        actions = self.get_actions(self.node.special_actions, menu)
        for a in actions:  # menu needed for 'parent'
            if type(a) == NodeItemAction:
                menu.addAction(a)
            elif type(a) == QMenu:
                menu.addMenu(a)

        return menu

    def itemChange(self, change, value):
        """Ensures that all connections, selection borders etc. that get drawn in the FlowView get
        constantly redrawn during a drag of the item"""

        if change == QGraphicsItem.ItemPositionChange:
            if self.session_design.performance_mode == 'pretty':
                self.flow_view.viewport().update()
            if self.movement_state == MovementEnum.mouse_clicked:
                self.movement_state = MovementEnum.position_changed

        self.update_conn_pos()

        return QGraphicsItem.itemChange(self, change, value)

    def update_conn_pos(self):
        """Updates the scene positions of connections"""

        for o in self.node.outputs:
            for c in o.connections:
                # c.item.recompute()

                if c not in self.flow_view.connection_items:
                    # it can happen that the connection item hasn't been
                    # created yet
                    continue

                item = self.flow_view.connection_items[c]
                item.recompute()
        for i in self.node.inputs:
            for c in i.connections:
                # c.item.recompute()

                if c not in self.flow_view.connection_items:
                    # it can happen that the connection item hasn't been
                    # created yet
                    continue

                item = self.flow_view.connection_items[c]
                item.recompute()

    def hoverEnterEvent(self, event):
        self.hovered = True
        self.widget.title_label.set_NI_hover_state(hovering=True)
        QGraphicsItem.hoverEnterEvent(self, event)

    def hoverLeaveEvent(self, event):
        self.hovered = False
        self.widget.title_label.set_NI_hover_state(hovering=False)
        QGraphicsItem.hoverLeaveEvent(self, event)

    def mousePressEvent(self, event):
        """Used for Moving-Commands in FlowView - may be replaced later by a nicer determination of a moving action."""

        self.flow_view.mouse_event_taken = True

        if event.button() == Qt.LeftButton:
            self.movement_state = MovementEnum.mouse_clicked
            self.movement_pos_from = self.pos()
        return QGraphicsItem.mousePressEvent(self, event)

    def mouseReleaseEvent(self, event):
        """Used for Moving-Commands in FlowView - may be replaced later by a nicer determination of a moving action."""

        self.flow_view.mouse_event_taken = True

        if self.movement_state == MovementEnum.position_changed:
            self.flow_view.selected_components_moved(self.pos() - self.movement_pos_from)
        self.movement_state = None
        return QGraphicsItem.mouseReleaseEvent(self, event)

    # ACTIONS

    def get_actions(self, actions_dict, menu):
        actions = []

        for k in actions_dict:
            v_dict = actions_dict[k]
            try:
                method = v_dict['method']
                data = None
                try:
                    data = v_dict['data']
                except KeyError:
                    pass
                action = NodeItemAction(node=self.node, text=k, method=method, menu=menu, data=data)
                action.triggered_with_data.connect(self.flow_view.thread_interface.trigger_node_action)
                action.triggered_without_data.connect(self.flow_view.thread_interface.trigger_node_action)

                actions.append(action)
            except KeyError:
                action_menu = QMenu(k, menu)
                sub_actions = self.get_actions(v_dict, action_menu)
                for a in sub_actions:
                    action_menu.addAction(a)
                actions.append(action_menu)

        return actions

    # CONFIG

    def complete_config(self, node_config):
        """
        Completes the node's config by all frontend related data
        """

        # add input widgets config
        for i in range(len(node_config['inputs'])):
            input_cfg = node_config['inputs'][i]
            inp_item = self.inputs[i]

            if inp_item.port.type_ == 'data':
                if inp_item.widget:
                    input_cfg['has widget'] = True

                    if inp_item.port.dtype:  # dtype widget
                        # input_cfg['widget name'] = str(inp_item.widget)
                        pass
                    else:  # custom widget
                        input_cfg['widget name'] = inp_item.port.add_config['widget name']
                        input_cfg['widget pos'] = inp_item.port.add_config['widget pos']

                    input_cfg['widget data'] = serialize(inp_item.widget.get_state())
                else:
                    input_cfg['has widget'] = False

                node_config['inputs'][i] = input_cfg

        # add item properties
        node_config['pos x'] = self.pos().x()
        node_config['pos y'] = self.pos().y()
        if self.main_widget:
            node_config['main widget data'] = serialize(self.main_widget.get_state())

        node_config['unconnected ports hidden'] = self.hiding_unconnected_ports
        node_config['collapsed'] = self.collapsed

        return node_config



from qtpy.QtWidgets import QGraphicsGridLayout, QGraphicsWidget, \
    QGraphicsLayoutItem
from qtpy.QtCore import Qt, QRectF, QPointF, QSizeF
from qtpy.QtGui import QFontMetricsF, QFont

from .PortItemInputWidgets import RCIW_BUILTIN_SpinBox, RCIW_BUILTIN_LineEdit, RCIW_BUILTIN_LineEdit_small, \
    Data_IW_S, Data_IW_M, Data_IW_L, Float_IW, Integer_IW, Choice_IW, Boolean_IW, String_IW_S, String_IW_M, String_IW_L
from .ryvencore import dtypes, FlowAlg
from .ryvencore.tools import deserialize
from .tools import get_longest_line, shorten

from .FlowViewProxyWidget import FlowViewProxyWidget


class PortItem(QGraphicsWidget):
    """The GUI representative for ports of nodes, also handling mouse events for connections."""

    def __init__(self, node, node_item, port, flow_view):
        super(PortItem, self).__init__()
        self.setGraphicsItem(self)

        self.node = node
        self.node_item = node_item
        self.port = port
        self.flow_view = flow_view

        # self.port.has_been_connected.connect(self.port_connected)
        # self.port.has_been_disconnected.connect(self.port_disconnected)

        self.pin = PortItemPin(self.port, self, self.node, self.node_item)

        self.label = PortItemLabel(self.port, self, self.node, self.node_item)

        self._layout = QGraphicsGridLayout()
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

    # --------------------------------------------
    def boundingRect(self):
        return QRectF(QPointF(0, 0), self.geometry().size())

    def setGeometry(self, rect):
        self.prepareGeometryChange()
        QGraphicsLayoutItem.setGeometry(self, rect)
        self.setPos(rect.topLeft())
    # --------------------------------------------

    def setup_ui(self):
        pass

    def port_connected(self):
        pass

    def port_disconnected(self):
        pass


class InputPortItem(PortItem):
    def __init__(self, node, node_item, port):
        super().__init__(node, node_item, port, node.flow)

        self.proxy = None  # widget proxy
        self.widget = self.create_widget()

        if self.widget:
            self.proxy = FlowViewProxyWidget(self.flow_view, parent=self.node_item)
            self.proxy.setWidget(self.widget)

        self.update_widget_value = self.widget is not None  # modified by FlowView when performance mode changes

        # catch up to missed connections
        if len(self.port.connections) > 0:
            self.port_connected()

        if self.port.add_config:

            if self.port.dtype:
                c_d = self.port.add_config['widget data']
                self.widget.set_state(deserialize(c_d))

            elif 'widget data' in self.port.add_config:
                try:
                    c_d = self.port.add_config['widget data']
                    if type(c_d) == dict:  # backwards compatibility
                        self.widget.set_state(c_d)
                    else:
                        self.widget.set_state(deserialize(c_d))
                except Exception as e:
                    print('Exception while setting data in', self.node.title,
                          '\'s input widget:', e, ' (was this intended?)')

        self.setup_ui()

    def setup_ui(self):
        l = self._layout

        # l.setSpacing(0)
        l.addItem(self.pin, 0, 0)
        l.setAlignment(self.pin, Qt.AlignVCenter | Qt.AlignLeft)
        l.addItem(self.label, 0, 1)
        l.setAlignment(self.label, Qt.AlignVCenter | Qt.AlignLeft)
        if self.widget:
            if self.port.add_config and self.port.add_config.get('widget pos') == 'below':
                l.addItem(self.proxy, 1, 0, 1, 2)
            else:
                l.addItem(self.proxy, 0, 2)  # besides

            l.setAlignment(self.proxy, Qt.AlignCenter)

    def create_widget(self):

        params = (self.port, self, self.node, self.node_item)

        if self.port.dtype:

            dtype = self.port.dtype

            if isinstance(dtype, dtypes.Data):
                if dtype.size == 's':
                    return Data_IW_S(params)
                elif dtype.size == 'm':
                    return Data_IW_M(params)
                elif dtype.size == 'l':
                    return Data_IW_L(params)

            elif isinstance(dtype, dtypes.String):
                if dtype.size == 's':
                    return String_IW_S(params)
                elif dtype.size == 'm':
                    return String_IW_M(params)
                elif dtype.size == 'l':
                    return String_IW_L(params)

            elif isinstance(dtype, dtypes.Integer):
                return Integer_IW(params)

            elif isinstance(dtype, dtypes.Float):
                return Float_IW(params)

            elif isinstance(dtype, dtypes.Boolean):
                return Boolean_IW(params)

            elif isinstance(dtype, dtypes.Choice):
                return Choice_IW(params)

        elif self.port.type_ == 'data' and self.port.add_config and 'widget name' in self.port.add_config:

            # custom input widget
            return self.get_input_widget_class(self.port.add_config['widget name'])(params)

        else:
            return None


    def get_input_widget_class(self, widget_name):
        """Returns the CLASS of a defined custom input widget"""
        return self.node.input_widget_classes[widget_name]

    def port_connected(self):
        """Disables the widget"""
        if self.widget:
            self.widget.setEnabled(False)

        if self.port.type_ == 'data':
            self.port.connections[0].activated.connect(self._port_val_updated)

        self._port_val_updated(self.port.val)

    def port_disconnected(self):
        """Enables the widget again"""
        if self.widget:
            self.widget.setEnabled(True)

    def _port_val_updated(self, val):
        """Called from output port"""

        if self.update_widget_value:  # this might be quite slow
            self.widget.val_update_event(val)


class OutputPortItem(PortItem):
    def __init__(self, node, node_item, port):
        super().__init__(node, node_item, port, node.flow)
        # super(OutputPortItem, self).__init__(parent_node_instance, PortObjPos.OUTPUT, type_, label_str)

        self.setup_ui()

    def setup_ui(self):
        l = self._layout

        # l.setSpacing(5)
        l.addItem(self.label, 0, 0)
        l.setAlignment(self.label, Qt.AlignVCenter | Qt.AlignRight)
        l.addItem(self.pin, 0, 1)

        l.setAlignment(self.pin, Qt.AlignVCenter | Qt.AlignRight)


# CONTENTS -------------------------------------------------------------------------------------------------------------


class PortItemPin(QGraphicsWidget):
    def __init__(self, port, port_item, node, node_item):
        super(PortItemPin, self).__init__(node_item)

        self.port = port
        self.port_item = port_item
        self.node = node
        self.node_item = node_item
        self.flow_view = self.node_item.flow_view

        self.setGraphicsItem(self)
        self.setAcceptHoverEvents(True)
        self.hovered = False
        self.setCursor(Qt.CrossCursor)
        self.tool_tip_pos = None

        self.padding = 2
        self.painting_width = 15
        self.painting_height = 15
        self.width = self.painting_width+2*self.padding
        self.height = self.painting_height+2*self.padding
        self.port_local_pos = None

    def boundingRect(self):
        return QRectF(QPointF(0, 0), self.geometry().size())

    def setGeometry(self, rect):
        self.prepareGeometryChange()
        QGraphicsLayoutItem.setGeometry(self, rect)
        self.setPos(rect.topLeft())

    def sizeHint(self, which, constraint=...):
        return QSizeF(self.width, self.height)

    def paint(self, painter, option, widget=None):
        self.node_item.session_design.flow_theme.paint_PI(
            node=self.node,
            painter=painter,
            option=option,
            node_color=self.node_item.color,
            type_=self.port.type_,
            connected=len(self.port.connections) > 0,
            padding=self.padding,
            w=self.painting_width,
            h=self.painting_height
        )

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:  # DRAG NEW CONNECTION
            self.flow_view.mouse_event_taken = True
            self.flow_view._selected_pin = self
            self.flow_view._dragging_connection = True
            event.accept()  # don't pass the ev ent to anything below
        else:
            return QGraphicsWidget.mousePressEvent(self, event)


    def hoverEnterEvent(self, event):
        if self.port.type_ == 'data':  # and self.parent_port_instance.io_pos == PortPos.OUTPUT:
            self.setToolTip(shorten(str(self.port.val), 1000, line_break=True))

        # update all connections
        for c in self.port.connections:
            conn_item = self.flow_view.connection_items[c]
            conn_item.update()

        self.hovered = True

        QGraphicsWidget.hoverEnterEvent(self, event)

    def hoverLeaveEvent(self, event):

        # turn connection highlighting off
        for c in self.port.connections:
            conn_item = self.flow_view.connection_items[c]
            conn_item.update()
        self.hovered = False

        QGraphicsWidget.hoverLeaveEvent(self, event)

    def get_scene_center_pos(self):
        if not self.node_item.collapsed:
            return QPointF(self.scenePos().x() + self.boundingRect().width()/2,
                           self.scenePos().y() + self.boundingRect().height()/2)
        else:
            if isinstance(self.port_item, InputPortItem):
                return self.node_item.get_left_body_header_vertex_scene_pos()
            else:
                return self.node_item.get_right_body_header_vertex_scene_pos()


class PortItemLabel(QGraphicsWidget):
    def __init__(self, port, port_item, node, node_item):
        super(PortItemLabel, self).__init__(node_item)
        self.setGraphicsItem(self)

        self.port = port
        self.port_item = port_item
        self.node = node
        self.node_item = node_item

        self.font = QFont("Source Code Pro", 10, QFont.Bold)
        font_metrics = QFontMetricsF(self.font)  # approximately! the designs can use different fonts
        self.width = font_metrics.width(get_longest_line(self.port.label_str))
        self.height = font_metrics.height() * (self.port.label_str.count('\n') + 1)
        self.port_local_pos = None

    def boundingRect(self):
        return QRectF(QPointF(0, 0), self.geometry().size())

    def setGeometry(self, rect):
        self.prepareGeometryChange()
        QGraphicsLayoutItem.setGeometry(self, rect)
        self.setPos(rect.topLeft())

    def sizeHint(self, which, constraint=...):
        return QSizeF(self.width, self.height)

    def paint(self, painter, option, widget=None):
        self.node_item.session_design.flow_theme.paint_PI_label(
            self.node,
            painter, option,
            self.port.type_,
            len(self.port.connections) > 0,
            self.port.label_str,
            self.node_item.color,
            self.boundingRect()
        )



from .Base import Base

from .RC import PortObjPos, FlowAlg
from .dtypes import DType
from .tools import serialize
from .InfoMsgs import InfoMsgs


class NodePort(Base):
    """The base class for inputs and outputs of nodes with basic functionality."""

    def __init__(self, node, io_pos, type_, label_str):
        Base.__init__(self)

        self.val = None
        self.node = node
        self.io_pos = io_pos
        self.type_ = type_
        self.label_str = label_str
        self.connections = []


    def get_val(self):
        pass

    def connected(self):
        pass

    def disconnected(self):
        pass

    def config_data(self):
        data_dict = {
            'type': self.type_,
            'label': self.label_str
        }

        return data_dict


class NodeInput(NodePort):

    def __init__(self, node, type_, label_str='', add_config=None, dtype: DType = None):
        super().__init__(node, PortObjPos.INPUT, type_, label_str)

        # add_config can be used to store additional config data for enhanced data input ports
        self.add_config = add_config

        # optional dtype
        self.dtype: DType = dtype

    def disconnected(self):
        super().disconnected()
        if self.type_ == 'data' and self.node.flow.alg_mode == FlowAlg.DATA:
            self.node.update(self.node.inputs.index(self))

    def get_val(self):
        InfoMsgs.write('getting value of node input')

        if self.node.flow.alg_mode == FlowAlg.DATA or len(self.connections) == 0:
            return self.val
        else:  # len(self.connections) > 0:
            return self.connections[0].get_val()

    def update(self, data=None):
        """called from another node or from connected()"""
        if self.type_ == 'data':
            self.val = data  # self.get_val()
            InfoMsgs.write('Data in input set to', data)

        self.node.update(inp=self.node.inputs.index(self))

    def config_data(self):
        data = super().config_data()

        if len(self.connections) == 0:
            data['val'] = serialize(self.get_val())

        if self.dtype:
            data['dtype'] = str(self.dtype)
            data['dtype state'] = serialize(self.dtype.get_state())

        return data



class NodeOutput(NodePort):
    def __init__(self, node, type_, label_str=''):
        super().__init__(node, PortObjPos.OUTPUT, type_, label_str)

    def exec(self):
        for c in self.connections:
            c.activate()

    def get_val(self):
        InfoMsgs.write('getting value in node output')

        if self.node.flow.alg_mode == FlowAlg.EXEC:
            self.node.update()
        return self.val

    def set_val(self, val):

        # in case val isn't of complex type
        self.val = val

        if self.node.flow.alg_mode == FlowAlg.DATA:
            for c in self.connections:
                c.activate(data=val)

    def connected(self):
        super().connected()
        if self.type_ == 'data' and self.node.flow.alg_mode == FlowAlg.DATA:
            self.set_val(self.val)  # update output



from qtpy.QtGui import QFont
from qtpy.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QLabel, QPushButton, QPlainTextEdit

from ..WRAPPERS import Logger
import logging


class LogWidget(QWidget):
    """Convenience class for a QWidget representing a log."""

    def __init__(self, logger: Logger):
        super().__init__()

        self.logger = logger
        self.logger.addHandler(logging.StreamHandler(self))
        self.logger.sig_disabled.connect(self.disable)
        self.logger.sig_enabled.connect(self.enable)

        self.main_layout = QVBoxLayout()
        self.header_layout = QHBoxLayout()

        title_label = QLabel(self.logger.name)
        title_label.setFont(QFont('Poppins', 12))
        self.header_layout.addWidget(title_label)

        self.remove_button = QPushButton('x')
        self.remove_button.clicked.connect(self.remove_clicked)
        self.header_layout.addWidget(self.remove_button)
        self.remove_button.hide()

        self.text_edit = QPlainTextEdit()
        self.text_edit.setReadOnly(True)

        self.main_layout.addLayout(self.header_layout)
        self.main_layout.addWidget(self.text_edit)

        self.setLayout(self.main_layout)

    def write(self, msg: str):
        self.text_edit.appendPlainText(msg)

    def flush(self):
        pass

    # def clear(self):
    #     self.text_edit.clear()

    def disable(self):
        self.remove_button.show()

    def enable(self):
        self.remove_button.hide()
        self.show()

    def remove_clicked(self):
        self.hide()



from .Base import Base

from .InfoMsgs import InfoMsgs


class Connection(Base):
    """
    The base class for both types of abstract connections. All data is transmitted through a connection from an output
    port to some connected input port. The classes ExecConnection and DataConnection are ready for reimplementation,
    so users can add additional functionality to connections (like "weights").
    Custom reimplementations must be provided in the RC.CLASSES dict.
    """

    def __init__(self, params):
        Base.__init__(self)

        self.out, self.inp, self.flow = params

    def activate(self):
        """Causes forward propagation of information"""
        pass



class ExecConnection(Connection):

    def activate(self):
        """Causes an update in the input port"""
        InfoMsgs.write('exec connection activated')

        self.inp.update()


class DataConnection(Connection):

    def __init__(self, params):
        super().__init__(params)

        self.data = None

    def get_val(self):
        """Gets the value of the output port -- only used in exec mode flows"""
        InfoMsgs.write('data connection getting value')

        # request data backwards
        self.data = self.out.get_val()

        return self.data

    def activate(self, data=None):
        """Passes data to the input port and causes update"""
        InfoMsgs.write('data connection activated')

        # store data
        self.data = data

        # propagate data forward
        self.inp.update(data)



from qtpy.QtWidgets import QGraphicsItem
from qtpy.QtGui import QPen, QPainter, QColor, QPainterPath
from qtpy.QtCore import Qt, QRectF, QPointF, QLineF

from .tools import MovementEnum


class DrawingObject(QGraphicsItem):
    """GUI implementation for 'drawing objects' in the scene, written by hand using a stylus pen"""

    def __init__(self, flow_view, config=None):
        super(DrawingObject, self).__init__()

        self.setFlags(QGraphicsItem.ItemIsSelectable | QGraphicsItem.ItemIsMovable |
                      QGraphicsItem.ItemSendsScenePositionChanges)
        self.setAcceptHoverEvents(True)
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)  # for rendering flow pictures

        self.flow_view = flow_view
        self.color = None
        self.base_stroke_weight = None
        self.type = 'pen'  # so far the only available, but I already save it so I could add more types in the future
        self.points = []
        self.stroke_weights = []
        self.pen_stroke_weight = 0  # approx. avg of self.stroke_weights
        self.rect = None
        self.path: QPainterPath = None
        self.width = -1
        self.height = -1
        self.finished = False

        # viewport_pos enables global floating points for precise pen positions
        self.viewport_pos: QPointF = config['viewport pos'] if 'viewport pos' in config else None
        # if the drawing gets loaded, its correct global floating pos is already correct (gets set by flow then)

        self.movement_state = None  # ugly - should get replaced later, see NodeItem, same issue
        self.movement_pos_from = None

        if 'points' in config:
            p_c = config['points']
            for p in p_c:
                if type(p) == list:
                    x = p[0]
                    y = p[1]
                    w = p[2]
                    self.points.append(QPointF(x, y))
                    self.stroke_weights.append(w)
                elif type(p) == dict:  # backwards compatibility
                    x = p['x']
                    y = p['y']
                    w = p['w']
                    self.points.append(QPointF(x, y))
                    self.stroke_weights.append(w)
            self.finished = True

        self.color = QColor(config['color'])
        self.base_stroke_weight = config['base stroke weight']


    def paint(self, painter, option, widget=None):

        if not self.finished:
            for i in range(1, len(self.points)):
                pen = QPen()
                pen.setColor(self.color)
                pen_width = (self.stroke_weights[i] + 0.2) * self.base_stroke_weight
                pen.setWidthF(pen_width)
                if i == 1 or i == len(self.points) - 1:
                    pen.setCapStyle(Qt.RoundCap)
                painter.setPen(pen)
                painter.setRenderHint(QPainter.Antialiasing)
                painter.setRenderHint(QPainter.HighQualityAntialiasing)
                painter.drawLine(self.points[i - 1], self.points[i])
            return

        if not self.path and self.finished:
            if len(self.points) == 0:
                return

            self.path = QPainterPath()
            self.path.moveTo(self.points[0])
            avg_weight = self.stroke_weights[0]
            for i in range(1, len(self.points)):
                self.path.lineTo(self.points[i])
                avg_weight += self.stroke_weights[i]
            self.pen_stroke_weight = (avg_weight/len(self.points) + 0.2)*self.base_stroke_weight

        pen = QPen()
        pen.setColor(self.color)
        pen.setWidthF(self.pen_stroke_weight)
        painter.setPen(pen)
        painter.setRenderHint(QPainter.Antialiasing)
        # painter.setRenderHint(QPainter.HighQualityAntialiasing)
        painter.drawPath(self.path)

    def append_point(self, posF_in_view: QPointF) -> bool:
        """
        Only used for active drawing.
        Appends a point (floating, in viewport coordinates) only if the distance to the last one isn't too small
        """

        p: QPointF = (self.viewport_pos + posF_in_view) - self.pos()
        p.setX(round(p.x(), 2))
        p.setY(round(p.y(), 2))

        if len(self.points) > 0:
            line = QLineF(self.points[-1], p)
            if line.length() < 0.5:
                return False

        self.points.append(p)
        return True

    def finish(self):
        """
        Computes the correct center position and updates the relative position for all points.
        """

        # Correct bounding rect (so far (0,0) is at the start of the line, but it should be in the middle)

        rect_center = self.get_points_rect_center()
        for p in self.points:
            p.setX(p.x()-rect_center.x())
            p.setY(p.y()-rect_center.y())
        self.setPos(self.pos()+rect_center)

        self.rect = self.get_points_rect()

        self.finished = True

    def get_points_rect(self):
        """Computes the 'bounding rect' for all points"""

        if len(self.points) == 0:
            return QRectF(0, 0, 0, 0)
        x_coords = [p.x() for p in self.points]
        y_coords = [p.y() for p in self.points]
        left = min(x_coords)
        right = max(x_coords)
        up = min(y_coords)
        down = max(y_coords)

        rect = QRectF(left, up, right - left, down - up)

        self.width = rect.width()
        self.height = rect.height()
        return rect

    def get_points_rect_center(self):
        """Returns the center point for the 'bounding rect' for all points"""

        return self.get_points_rect().center()

    def boundingRect(self):
        if self.rect:
            return self.rect
        else:
            return self.get_points_rect()

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            self.flow_view.viewport().update()
            if self.movement_state == MovementEnum.mouse_clicked:
                self.movement_state = MovementEnum.position_changed

        return QGraphicsItem.itemChange(self, change, value)

    def mousePressEvent(self, event):
        """Used for Moving-Commands in Flow - may be replaced later with a nicer determination of a move action."""

        self.movement_state = MovementEnum.mouse_clicked
        self.movement_pos_from = self.pos()
        return QGraphicsItem.mousePressEvent(self, event)

    def mouseReleaseEvent(self, event):
        """Used for Moving-Commands in Flow - may be replaced later with a nicer determination of a move action."""

        if self.movement_state == MovementEnum.position_changed:
            self.flow_view.selected_components_moved(self.pos() - self.movement_pos_from)
        self.movement_state = None
        return QGraphicsItem.mouseReleaseEvent(self, event)

    def config_data(self):
        drawing_dict = {
            'pos x': self.pos().x(),
            'pos y': self.pos().y(),
            'color': self.color.name(),
            'type': self.type,
            'base stroke weight': self.base_stroke_weight
        }
        points_list = []
        for i in range(len(self.points)):
            p = self.points[i]
            points_list.append([p.x(), p.y(), self.stroke_weights[i]])
        drawing_dict['points'] = points_list
        return drawing_dict



"""
This file contains the implementations of the possible undoable actions for FlowView.

Why not in Flow?
-> Because there are actions that only regard the GUI and are therefore unnoticed by the Flow but still undoable, like
adding a DrawingObject.

The main characteristic of these implementations is that they communicate with the Flow only using signals and slots,
which complicates the process of making changes but enables threading, as in case of the abstract
components being managed in a separate thread, no direct communication between the FlowCommands from the main thread
and the Flow are allowed. Usually to make changes to the flow, they emit signals that are temporarily connected
the the Flow which then causes the FlowView to update accordingly.

These implementations might have quite some potential for improvement.
"""


from qtpy.QtCore import Signal, QObject, QPointF
from qtpy.QtWidgets import QUndoCommand

from .DrawingObject import DrawingObject
from .ryvencore.Node import Node
from .ryvencore.Connection import Connection
from .NodeItem import NodeItem
from .ryvencore.NodePort import NodePort


class FlowUndoCommand(QUndoCommand, QObject):
    """
    The main difference to normal QUndoCommands is the activation feature. This allows the flow widget to add the
    undo command to the undo stack before redo() is called. This is important since some of these commands can cause
    other commands to be added while they are performing redo(), so to prevent those commands to be added to the
    undo stack before the parent command, it is here blocked at first.
    """

    def __init__(self, flow_view):

        self.flow_view = flow_view
        self.flow = flow_view.flow
        self._activated = False

        QUndoCommand.__init__(self)
        QObject.__init__(self)

    def activate(self):
        self._activated = True
        self.redo()

    def redo(self) -> None:
        if not self._activated:
            return
        else:
            self.redo_()

    def undo(self) -> None:
        self.undo_()

    def redo_(self):
        """subclassed"""
        pass

    def undo_(self):
        """subclassed"""
        pass


class MoveComponents_Command(FlowUndoCommand):
    def __init__(self, flow_view, items_list, p_from, p_to):
        super(MoveComponents_Command, self).__init__(flow_view)

        self.items_list = items_list
        self.p_from = p_from
        self.p_to = p_to
        self.last_item_group_pos = p_to

    def undo_(self):
        items_group = self.items_group()
        items_group.setPos(self.p_from)
        self.last_item_group_pos = items_group.pos()
        self.destroy_items_group(items_group)

    def redo_(self):
        items_group = self.items_group()
        items_group.setPos(self.p_to - self.last_item_group_pos)
        self.destroy_items_group(items_group)


    def items_group(self):
        return self.flow_view.scene().createItemGroup(self.items_list)

    def destroy_items_group(self, items_group):
        self.flow_view.scene().destroyItemGroup(items_group)


class PlaceNode_Command(FlowUndoCommand):

    create_node_request = Signal(object)
    add_node_request = Signal(Node)
    remove_node_request = Signal(Node)

    def __init__(self, flow_view, node_class, pos):
        super().__init__(flow_view)

        self.node_class = node_class
        self.node = None
        self.item_pos = pos

        self.create_node_request.connect(self.flow.create_node)
        self.add_node_request.connect(self.flow.add_node)
        self.remove_node_request.connect(self.flow.remove_node)

        self.flow_view.node_placed.connect(self.node_placed_in_flow)

    def undo_(self):
        self.remove_node_request.emit(self.node)

    def redo_(self):
        if self.node:
            self.add_node_request.emit(self.node)
        else:
            self.flow_view.node_placed.connect(self.node_placed_in_flow)
            self.create_node_request.emit(self.node_class)

        # --> node_placed_in_flow()

    def node_placed_in_flow(self, node):
        self.flow_view.node_placed.disconnect(self.node_placed_in_flow)
        self.node = node


class PlaceDrawing_Command(FlowUndoCommand):
    def __init__(self, flow_view, posF, drawing):
        super().__init__(flow_view)

        self.drawing = drawing
        self.drawing_obj_place_pos = posF
        self.drawing_obj_pos = self.drawing_obj_place_pos

    def undo_(self):
        # The drawing_obj_pos is not anymore the drawing_obj_place_pos because after the
        # drawing object was completed, its actual position got recalculated according to all points and differs from
        # the initial pen press pos (=drawing_obj_place_pos). See DrawingObject.finished().

        self.drawing_obj_pos = self.drawing.pos()

        self.flow_view.remove_component(self.drawing)

    def redo_(self):
        self.flow_view.add_drawing(self.drawing, self.drawing_obj_pos)


class RemoveComponents_Command(FlowUndoCommand):

    add_node_request = Signal(Node)
    remove_node_request = Signal(Node)

    add_connection_request = Signal(Connection)
    remove_connection_request = Signal(Connection)

    def __init__(self, flow_view, items):
        super().__init__(flow_view)

        self.items = items
        self.broken_connections = []  # the connections that go beyond the removed nodes and need to be restored in undo
        self.internal_connections = set()

        # static connections
        self.add_node_request.connect(self.flow.add_node)
        self.remove_node_request.connect(self.flow.remove_node)
        self.add_connection_request.connect(self.flow.add_connection)
        self.remove_connection_request.connect(self.flow.remove_connection)

        self.node_items = []
        self.nodes = []
        self.drawings = []
        for i in self.items:
            if isinstance(i, NodeItem):
                self.node_items.append(i)
                self.nodes.append(i.node)
            elif isinstance(i, DrawingObject):
                self.drawings.append(i)

        for n in self.nodes:
            for i in n.inputs:
                for c in i.connections:
                    cp = c.out
                    cn = cp.node
                    if cn not in self.nodes:
                        self.broken_connections.append(c)
                    else:
                        self.internal_connections.add(c)
            for o in n.outputs:
                for c in o.connections:
                    cp = c.inp
                    cn = cp.node
                    if cn not in self.nodes:
                        self.broken_connections.append(c)
                    else:
                        self.internal_connections.add(c)

    def undo_(self):
        # add connections
        self.restore_broken_connections()
        self.restore_internal_connections()

        # add nodes
        for n in self.nodes:
            self.add_node_request.emit(n)

        # add drawings
        for d in self.drawings:
            self.flow_view.add_drawing(d)

    def redo_(self):

        # remove connections
        self.remove_broken_connections()
        self.remove_internal_connections()

        # remove nodes
        for n in self.nodes:
            self.remove_node_request.emit(n)

        # remove drawings
        for d in self.drawings:
            self.flow_view.remove_drawing(d)

    def restore_internal_connections(self):
        for c in self.internal_connections:
            self.add_connection_request.emit(c)

    def remove_internal_connections(self):
        for c in self.internal_connections:
            self.remove_connection_request.emit(c)

    def restore_broken_connections(self):
        for c in self.broken_connections:
            self.add_connection_request.emit(c)

    def remove_broken_connections(self):
        for c in self.broken_connections:
            self.remove_connection_request.emit(c)


class ConnectPorts_Command(FlowUndoCommand):

    connect_request = Signal(NodePort, NodePort)
    add_connection_request = Signal(Connection)
    remove_connection_request = Signal(Connection)

    def __init__(self, flow_view, out, inp):
        super().__init__(flow_view)

        # CAN ALSO LEAD TO DISCONNECT INSTEAD OF CONNECT!!

        self.out = out
        self.inp = inp
        self.connection = None
        self.connecting = True

        # static connections
        self.connect_request.connect(self.flow.connect_nodes)
        self.add_connection_request.connect(self.flow.add_connection)
        self.remove_connection_request.connect(self.flow.remove_connection)

        for c in self.out.connections:
            if c.inp == self.inp:
                self.connection = c
                self.connecting = False


    def undo_(self):
        if self.connecting:
            # remove connection
            self.remove_connection_request.emit(self.connection)
        else:
            # recreate former connection
            self.add_connection_request.emit(self.connection)

    def redo_(self):
        if self.connecting:
            if self.connection:
                self.add_connection_request.emit(self.connection)
            else:
                # connection hasn't been created yet
                self.flow.connection_added.connect(self.connection_created)
                self.connect_request.emit(self.out, self.inp)
        else:
            # remove existing connection
            self.remove_connection_request.emit(self.connection)

    def connection_created(self, c):
        self.flow.connection_added.disconnect(self.connection_created)
        self.connection = c




class Paste_Command(FlowUndoCommand):

    create_nodes_request = Signal(list)
    create_connections_request = Signal(list, list)

    add_node_request = Signal(Node)
    remove_node_request = Signal(Node)

    add_connection_request = Signal(Connection)
    remove_connection_request = Signal(Connection)


    def __init__(self, flow_view, data, offset_for_middle_pos):
        super().__init__(flow_view)

        self.data = data
        self.modify_data_positions(offset_for_middle_pos)
        self.pasted_components = None

        # static connections
        self.add_node_request.connect(self.flow.add_node)
        self.remove_node_request.connect(self.flow.remove_node)
        self.add_connection_request.connect(self.flow.add_connection)
        self.remove_connection_request.connect(self.flow.remove_connection)


    def modify_data_positions(self, offset):
        """adds the offset to the components' positions in data"""

        for node in self.data['nodes']:
            node['pos x'] = node['pos x'] + offset.x()
            node['pos y'] = node['pos y'] + offset.y()
        for drawing in self.data['drawings']:
            drawing['pos x'] = drawing['pos x'] + offset.x()
            drawing['pos y'] = drawing['pos y'] + offset.y()


    def connect_to_flow(self):
        """creates temporary connections to retrieve the created components once"""
        self.flow.nodes_created_from_config.connect(self.nodes_created)
        self.flow.connections_created_from_config.connect(self.connections_created)
        self.create_nodes_request.connect(self.flow.create_nodes_from_config)
        self.create_connections_request.connect(self.flow.connect_nodes_from_config)

    def disconnect_from_flow(self):
        self.flow.nodes_created_from_config.disconnect(self.nodes_created)
        self.flow.connections_created_from_config.disconnect(self.connections_created)
        self.create_nodes_request.disconnect(self.flow.create_nodes_from_config)
        self.create_connections_request.disconnect(self.flow.connect_nodes_from_config)


    def redo_(self):
        if self.pasted_components is None:
            self.pasted_components = {}

            # create components
            self.create_drawings()
            self.connect_to_flow()
            self.create_nodes_request.emit(self.data['nodes'])
            # --> nodes_created()
        else:
            self.add_existing_components()

    def undo_(self):
        # remove components and their items from flow
        for n in self.pasted_components['nodes']:
            self.remove_node_request.emit(n)
        for c in self.pasted_components['connections']:
            self.remove_connection_request.emit(c)
        for d in self.pasted_components['drawings']:
            self.flow_view.remove_drawing(d)

    def add_existing_components(self):
        # add existing components and items to flow
        for n in self.pasted_components['nodes']:
            self.add_node_request.emit(n)
        for c in self.pasted_components['connections']:
            self.add_connection_request.emit(c)
        for d in self.pasted_components['drawings']:
            self.flow_view.add_drawing(d)

        self.select_new_components_in_view()

    def select_new_components_in_view(self):
        self.flow_view.clear_selection()
        for d in self.pasted_components['drawings']:
            d: DrawingObject
            d.setSelected(True)
        for n in self.pasted_components['nodes']:
            n: NodeItem
            ni: NodeItem = self.flow_view.node_items[n]
            ni.setSelected(True)




    def nodes_created(self, nodes):
        self.pasted_components['nodes'] = nodes
        self.create_connections_request.emit(nodes, self.data['connections'])
        # --> connections_created()

    def connections_created(self, connections):
        self.disconnect_from_flow()
        self.pasted_components['connections'] = connections

        # self.create_drawings()
        self.select_new_components_in_view()

    def create_drawings(self):
        drawings = []
        for d in self.data['drawings']:
            new_drawing = self.flow_view.create_drawing(d)
            self.flow_view.add_drawing(new_drawing, posF=QPointF(d['pos x'], d['pos y']))
            drawings.append(new_drawing)
        self.pasted_components['drawings'] = drawings



from .Base import Base

from .RC import CLASSES


class Script(Base):
    """A Script consists of a Flow, the corresponding FlowView, a variables manager and a logger."""

    def __init__(self, session, title: str = None, config_data: dict = None, create_default_logs=True):
        Base.__init__(self)

        self.session = session
        self.logs_manager = None
        self._create_default_logs = create_default_logs
        self.vars_manager = None
        self.title = title if not config_data else config_data['name']
        self.flow = None

        self.init_config = config_data
        self.init_flow_config = None
        self.init_vars_manager_config = None

        # saving config data
        if config_data:
            self.init_flow_config = config_data['flow'] if config_data else None
            self.init_vars_manager_config = config_data['variables']

        # logging
        self.logs_manager = CLASSES['logs manager'](self, self._create_default_logs)

        # vars manager
        self.vars_manager = CLASSES['vars manager'](self, self.init_vars_manager_config)

        # flow
        self.flow = CLASSES['flow'](self.session, self)

    def load_flow(self):
        if self.init_flow_config:
            self.flow.load(self.init_flow_config)


    def serialize(self) -> dict:
        """Returns the config data of the script, including variables and flow content"""

        flow_data = self.flow.generate_config_data()

        script_dict = {
            'name': self.title,
            'variables': self.vars_manager.config_data(),
            'flow': flow_data,
            # 'flow widget config': flow_view_config,
        }

        return script_dict
